"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import io

import requests

from ..formdata import FormData
from ..exceptions import *
from .base import Transport, RawResponse

import json

NOT_MODIFIED = 304

class RequestsTransport(Transport):
    """ This is a simple transport based on the requests library, it's not advised to use in production """
    def __init__(self, session, requests_session: requests.Session = None):
        super().__init__(session)
        
        self._s = requests_session or requests.Session()

    @classmethod
    def _get_priority(cls):
        try:
            return 3
        except ImportError:
            return None

    def _build_raw(self, ret):
        response = RawResponse(ret.status_code,
                               tuple(ret.headers.items()), None, None)

        if ret.status_code == NOT_MODIFIED:
            return response

        response.data = ret.content

        if ret.headers['content-type'] != 'application/octet-stream':
            response.json = self._parse_json(ret, allow_unmodified=True)

        return response

    def _parse_json(self, ret, allow_unmodified=False):
        if allow_unmodified and ret.status_code == NOT_MODIFIED:
            return None

        try:
            ret_json = ret.json()
        except json.decoder.JSONDecodeError:
            raise ProtonAPIError(ret.status_code, dict(ret.headers), {})

        if ret_json['Code'] not in [1000, 1001]:
            raise ProtonAPIError(ret.status_code, dict(ret.headers), ret_json)

        return ret_json

    async def async_api_request(
        self, endpoint,
        jsondata=None, data=None, additional_headers=None,
        method=None, params=None, return_raw=False
    ):
        self._s.headers['x-pm-appversion'] = self._session.appversion
        self._s.headers['User-Agent'] = self._session.user_agent

        if self._session.authenticated:
            self._s.headers['x-pm-uid'] = self._session.UID
            self._s.headers['Authorization'] = 'Bearer ' + self._session.AccessToken

        # If we don't have an explicit method, default to get if there's no data, post otherwise
        if method is None:
            if not jsondata and not data:
                fct = self._s.get
            else:
                fct = self._s.post
        else:
            fct = {
                'get': self._s.get,
                'post': self._s.post,
                'put': self._s.put,
                'delete': self._s.delete,
                'patch': self._s.patch
            }.get(method.lower())

            if fct is None:
                raise ValueError("Unknown method: {}".format(method))

        data_dict = self._get_requests_data(data) if data else None
        files_dict = self._get_requests_files(data) if data else None
        try:
            ret = fct(
                self._environment.http_base_url + endpoint,
                headers=additional_headers,
                json=jsondata,
                data=data_dict,
                files=files_dict,
                params=params
            )
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
            raise ProtonAPINotReachable(e)
        except (Exception, requests.exceptions.BaseHTTPError) as e:
            raise ProtonAPIUnexpectedError(e)

        if return_raw:
            return self._build_raw(ret)

        ret_json = self._parse_json(ret)

        return ret_json

    @staticmethod
    def _get_requests_data(form_data: FormData) -> dict:
        """
        Converts the FormData instance to a dict that can be passed
        as the data parameter in requests (e.g. `requests.post(url, data=data)`.

        File-like fields are ignored, use `_get_requests_files` for those.
        """
        return {
            field.name: field.value
            for field in form_data.fields if not isinstance(field.value, io.IOBase)
        }

    @staticmethod
    def _get_requests_files(form_data: FormData) -> dict:
        """
        Extracts the file-like fields to a dict that can be passed as the `files`
        parameter in requests (e.g. `requests.post(url, files=files`).
        """
        # From https://requests.readthedocs.io/en/latest/api/#requests.request:
        # files – (optional) Dictionary of 'name': file-like-objects
        # (or {'name': file-tuple}) for multipart encoding upload. file-tuple
        # can be a 2-tuple ('filename', fileobj), 3-tuple ('filename', fileobj, 'content_type')
        # or a 4-tuple ('filename', fileobj, 'content_type', custom_headers),
        # where 'content-type' is a string defining the content type of the
        # given file and custom_headers a dict-like object containing additional
        # headers to add for the file.
        return {
            field.name: (field.filename, field.value, field.content_type)
            for field in form_data.fields if isinstance(field.value, io.IOBase)
        }



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from importlib import metadata
import os
import inspect
import threading
import warnings
from collections import namedtuple
from typing import Optional

from ..utils import Singleton

PluggableComponent = namedtuple('PluggableComponent', ['priority', 'class_name', 'cls'])
PluggableComponentName = namedtuple('PluggableComponentName', ['type_name', 'class_name'])


class Loader(metaclass=Singleton):
    """This is the loader for pluggable components. These components are identified by a type name (string)
    and a class name (also a string).

    In normal use, one will only use :meth:`get`, as follows:

    .. code-block::

        from proton.loader import Loader
        # Note the parenthesis to instanciate an object, as Loader.get() returns a class.
        my_keyring = Loader.get('keyring')()

    You can influence which component to use using the ``PROTON_LOADER_OVERRIDES`` environment variable. It's a comma separated list
    of ``type_name=class_name`` (to force ``class_name`` to be used) and ``type_name=-class_name`` (to exclude ``class_name`` from the options considered).

    To find the candidates, ``Loader`` will use entry points, that are to be defined in setup.py, as follows:

    .. code-block::

        setup(
            #[...],
            entry_points={
                "proton_loader_keyring": [
                    "json = proton.keyring.textfile:KeyringBackendJsonFiles"
                ]
            },
            #[...]
        )

    The class pointed by these entrypoints should implement the following class methods:

    * :meth:`_get_priority`: return a numeric value, larger ones have higher priority. If it's ``None``, then this class won't be considered
    * :meth:`_validate`: check if the object can indeed be used (might be expensive/annoying). If it returns ``False``, then the backend won't be considered for the rest of the session.

    If :meth:`_validate` is not defined, then it's assumed that it will always succeed.

    To display the list of valid values, you can use ``python3 -m proton.loader``.

    """

    __loader_prefix = 'proton_loader_'

    def __init__(self):
        self.__known_types = {}
        self.__name_resolution_cache = {}
        self.__lock = threading.Lock()

    def get(
        self, type_name: str,
        class_name: Optional[str] = None,
        validate_params: Optional[dict] = None
    ) -> type:
        """Get the implementation for type_name.

        :param type_name: extension type
        :type type_name: str
        :param class_name: specific implementation to get, defaults to None (use preferred one)
        :type class_name: Optional[str], optional
        :param validate_params: pass custom arguments to backends during `_validate`
            Could be used in cases when the validity of a backend is to be dynamically
            evaluated upon calling `get()`
        :type validate_params: Optional[dict], optional
        :raises RuntimeError: if no valid implementation can be found,
            or if PROTON_LOADER_OVERRIDES is invalid.
        :return: the class implementing type_name. (careful: it's a class, not an object!)
        :rtype: class
        """
        acceptable_classes = self.get_all(type_name)

        for entry in acceptable_classes:
            # If caller specified the class he wanted, then we check only that.
            if class_name is not None:
                if entry.class_name == class_name:
                    return entry.cls
                else:
                    continue

            # Invalid priority, just continue
            # (this will fail anyway because we have ordered the list in get_all,
            # but for what it costs I prefer to go through the list)
            if entry.priority is None:
                continue

            # If we have a _validate class method, try to see if the object is indeed acceptable
            if hasattr(entry.cls, '_validate'):
                if inspect.signature(entry.cls._validate).parameters.get("validate_params"):  # noqa: E501  pylint: disable=protected-access
                    entry_is_valid = entry.cls._validate(validate_params)  # noqa: E501 pylint: disable=protected-access
                else:
                    entry_is_valid = entry.cls._validate()  # pylint: disable=protected-access

                if entry_is_valid:
                    return entry.cls
            else:
                return entry.cls

        raise RuntimeError(f"Loader: couldn't find an acceptable implementation for {type_name}.")

    @property
    def type_names(self) -> list[str]:
        """
        :return: Return a list of the known type names
        :rtype: list[str]
        """
        return [x[len(self.__loader_prefix):] for x in self._proton_entry_point_groups.keys()]

    @property
    def _proton_entry_point_groups(self):
        metadata_entry_points = metadata.entry_points()
        try:
            # importlib.metadata.entry_points() uses the selectable interface in python >= 3.10
            groups = metadata_entry_points.groups
        except AttributeError:
            # importlib.metadata.entry_points() uses the dict interface in python < 3.10
            return {
                k: v
                for k, v in metadata_entry_points.items()
                if k.startswith(self.__loader_prefix)
            }

        return {
            group: metadata_entry_points.select(group=group)
            for group in groups if group.startswith(self.__loader_prefix)
        }

    def get_all(self, type_name: str) -> list[PluggableComponent]:
        """Get a list of all implementations for ``type_name``.

        :param type_name: type of implementation to query for
        :type type_name: str
        :raises RuntimeError: if ``PROTON_LOADER_OVERRIDES`` has conflicts
        :return: Implementation for type_name (this includes the ones that are disabled)
        :rtype: list[PluggableComponent]
        """

        # If we don't have already loaded the entry points, just do so
        with self.__lock:
            # We use a lock here because a known type should only be available after it has been loaded.
            if type_name not in self.__known_types:
                metadata_group_name = self._get_metadata_group_for_typename(type_name)
                entry_points = self._proton_entry_point_groups.get(metadata_group_name, ())
                self.__known_types[type_name] = {}
                for ep in entry_points:
                    if ep.name in self.__known_types[type_name]:
                        del self.__known_types[type_name]
                        raise RuntimeError("Loader error : found 2 modules with same name (that would create security issues)")
                    try:
                        self.__known_types[type_name][ep.name] = ep.load()
                    except AttributeError:
                        warnings.warn(f"Loader: couldn't load {type_name}/{ep.name}, is it installed properly?", RuntimeWarning, stacklevel=2)
                        continue
                    self.__name_resolution_cache[self.__known_types[type_name][ep.name]] = PluggableComponentName(type_name, ep.name)

        # We do this at runtime, because we want to make sure we can change it after start.
        overrides = os.environ.get('PROTON_LOADER_OVERRIDES', '')
        overrides = [x.strip() for x in overrides.split()]
        overrides = [x[len(type_name)+1:] for x in overrides if x.startswith(f'{type_name}=')]

        force_class = set([x for x in overrides if not x.startswith('-')])
        if len(force_class) == 1:
            force_class = list(force_class)[0]
            if force_class in self.__known_types[type_name]:
                acceptable_entry_points = [force_class]
        elif len(force_class) > 1:
            raise RuntimeError(f"Loader: PROTON_LOADER_OVERRIDES contains multiple force for {type_name}")
        else:
            # Load all entry_points, except those that are excluded by PROTON_LOADER_OVERRIDES
            acceptable_entry_points = []
            for k in self.__known_types[type_name].keys():
                if '-' + k not in overrides:
                    acceptable_entry_points.append(k)

        acceptable_classes = [(v._get_priority(), k, v) for k, v in self.__known_types[type_name].items() if k in acceptable_entry_points]
        acceptable_classes += [(None, k, v) for k, v in self.__known_types[type_name].items() if k not in acceptable_entry_points]
        acceptable_classes_with_prio = [PluggableComponent(priority, class_name, v) for priority, class_name, v in acceptable_classes if priority is not None]
        acceptable_classes_without_prio = [PluggableComponent(priority, class_name, v) for priority, class_name, v in acceptable_classes if priority is None]

        # Sort the entries with priority, highest first
        acceptable_classes_with_prio.sort(reverse=True)

        return acceptable_classes_with_prio + acceptable_classes_without_prio

    def get_name(self, cls: type) -> Optional[PluggableComponentName]:
        """Return the type_name and class_name corresponding to the class in parameter.

        This is useful for inverse lookups (i.e. for logs for instance)

        :return: ``Tuple (type_name, class_name)``
        :rtype: Optional[PluggableComponentName]
        """
        return self.__name_resolution_cache.get(cls, None)

    def reset(self) -> None:
        """Erase the loader cache. (useful for tests)"""
        self.__known_types = {}
        self.__name_resolution_cache = {}

    def set_all(self, type_name: str, implementations: dict[str, type]):
        """Set a defined set of implementation for a given ``type_name``.

        This method is probably useful only for testing.

        :param type_name: Type
        :type type_name: str
        :param implementations: Dictionary implementation name -> implementation class
        :type implementations: dict[str, class]
        """
        self.__known_types[type_name] = implementations
        for class_name, cls in implementations.items():
            self.__name_resolution_cache[cls] = PluggableComponentName(type_name, class_name)

    def _get_metadata_group_for_typename(self, type_name: str) -> str:
        """Return the metadata group name for type_name

        :param type_name: type_name
        :type type_name: str
        :return: metadata group name
        :rtype: str
        """
        return self.__loader_prefix + type_name



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass
from typing import Awaitable, List
import aiohttp
from ..exceptions import *
from .aiohttp import AiohttpTransport

import json, base64, struct, time, asyncio, random, itertools

from urllib.parse import urlparse

from ..api import sync_wrapper
from .utils.dns import DNSParser, DNSResponseError


@dataclass
class AlternativeRoutingDNSQueryAnswer:
    """Contains the result of a successful DNS query to retrieve the
    alternative routing server domain."""
    expiration_time: float
    domain: str


class AlternativeRoutingTransport(AiohttpTransport):
    DNS_PROVIDERS = [
        #dns.google
        (("8.8.4.4", "8.8.8.8"), ("2001:4860:4860::8844", "2001:4860:4860::8888"), '/dns-query'),
        #dns11.quad9.net
        (("149.112.112.11", "9.9.9.11"), ("2620:fe::fe:11", "2620:fe::11"), '/dns-query'),
    ]

    STRUCT_REPLY_COUNTS = struct.Struct('>HHHH')
    STRUCT_REC_FORMAT = struct.Struct('>HHIH')

    #Delay between DNS requests
    DELAY_DNS_REQUEST = 2
    TIMEOUT_DNS_REQUEST = 10

    @classmethod
    def _get_priority(cls):
        return 5


    def __init__(self, session):
        super().__init__(session)
        self._alternative_routes = []

    @classmethod
    def _compute_ar_domain(cls, host):
        return b'd' + base64.b32encode(host.encode('ascii')).strip(b'=') + b".protonpro.xyz"

    async def _async_dns_query(
            self, domain, dns_server_ip, dns_server_path, delay=0
    ) -> List[AlternativeRoutingDNSQueryAnswer]:
        import aiohttp

        if delay > 0:
            await asyncio.sleep(delay)

        ardomain = self._compute_ar_domain(domain)
        dns_request = DNSParser.build_query(ardomain, qtype=16, qclass=1)  # TXT IN
        dot_url = f'https://{dns_server_ip}{dns_server_path}'

        async with aiohttp.ClientSession() as session:
            async with session.post(dot_url, headers=[("Content-Type","application/dns-message")], data=dns_request) as r:
                reply_data = await r.content.read()

        try:
            dns_answers = DNSParser.parse(reply_data)
        except DNSResponseError as e:
            raise ProtonAPINotReachable(str(e))

        now = time.time()
        # Tuples (TTL, data)
        answers = []
        for rec_ttl, rec_val in dns_answers:
            answers.append(AlternativeRoutingDNSQueryAnswer(
                expiration_time=now + rec_ttl,
                domain=rec_val)
            )

        return answers

    @property
    def _http_domain(self):
        return urlparse(super().http_base_url).netloc

    async def _get_alternative_routes(self):
        # We generate a random list of dns servers, 
        # we query them following that order, simultaneoulsy on IPv4/IPv6
        choices_ipv4 = []
        choices_ipv6 = []
        for dns_server_ipv4s, dns_server_ipv6s, dns_server_path in self.DNS_PROVIDERS:
            for ip in dns_server_ipv4s:
                choices_ipv4.append((ip, dns_server_path))
            for ip in dns_server_ipv6s:
                choices_ipv6.append((ip, dns_server_path))

        random.shuffle(choices_ipv4)
        random.shuffle(choices_ipv6)

        pending = []
        i = 0
        for ipv4, ipv6 in itertools.zip_longest(choices_ipv4, choices_ipv6, fillvalue=None):
            if i * self.DELAY_DNS_REQUEST > self.TIMEOUT_DNS_REQUEST:
                break

            if ipv4 is not None:
                pending.append(asyncio.create_task(self._async_dns_query(self._http_domain, ipv4[0], ipv4[1], delay=i * self.DELAY_DNS_REQUEST)))
            if ipv6 is not None:
                pending.append(asyncio.create_task(self._async_dns_query(self._http_domain, f'[{ipv6[0]}]', ipv6[1], delay=i * self.DELAY_DNS_REQUEST)))
            
            i += 1

        results_ok = []
        results_fail = []
        proton_api_not_available_errors = []
        final_timestamp = time.time() + self.TIMEOUT_DNS_REQUEST
        while len(pending) > 0 and len(results_ok) == 0:
            done, pending = await asyncio.wait(pending, timeout=max(0.1, final_timestamp - time.time()), return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                try:
                    results_ok += task.result()
                except ProtonAPINotAvailable as e:
                    # That means that we were able to do a resolution, but it explicitly failed
                    proton_api_not_available_errors.append(e)
                except Exception as e:
                    results_fail.append(e)
        
        for task in pending:
            task.cancel()

        if proton_api_not_available_errors:
            raise proton_api_not_available_errors[0]

        if len(results_ok) == 0:
            if len(self._alternative_routes) > 0:
                # We have routes, but we were not able to resolve new ones. Just keep the old ones
                return
            else:
                # No routes, and failed to get new ones
                raise ProtonAPINotReachable("Couldn't resolve any alternative routing names")

        domains = [x.domain for x in results_ok]
        # Filter names that are in our results (we don't want duplicates)
        self._alternative_routes = [
            x for x in self._alternative_routes
            if x.domain not in domains and x.expiration_time >= time.time()
        ]
        # Add the results
        self._alternative_routes += results_ok
        # Sort them so we have the most recent on top
        self._alternative_routes.sort(key=lambda x: x.expiration_time, reverse=True)

    @property
    def http_base_url(self):
        if len(self._alternative_routes) == 0:
            raise ProtonAPINotReachable("AlternativeRouting transport doesn't have any route")

        path = urlparse(super().http_base_url).path
        
        return f'https://{self._alternative_routes[0].domain}{path}'

    @property
    def tls_pinning_hashes(self):
        return self._environment.tls_pinning_hashes_ar

    async def async_api_request(
        self, endpoint,
        jsondata=None, data=None, additional_headers=None,
        method=None, params=None, return_raw=False
    ):
        if len(self._alternative_routes) == 0 or self._alternative_routes[0].expiration_time < time.time():
            await self._get_alternative_routes()

        return await super().async_api_request(endpoint, jsondata, data, additional_headers, method,
                                               params, return_raw=return_raw)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# Custom expanded version of SHA512
import hashlib


class PMHash:
    digest_size = 256
    name = 'PMHash'

    def __init__(self, b=b""):
        self.b = b

    def update(self, b):
        self.b += b

    def digest(self):
        return hashlib.sha512(
                self.b + b'\0'
            ).digest() + hashlib.sha512(
                self.b + b'\1'
            ).digest() + hashlib.sha512(
                self.b + b'\2'
            ).digest() + hashlib.sha512(
                self.b + b'\3'
            ).digest()

    def hexdigest(self):
        return self.digest().hex()

    def copy(self):
        return PMHash(self.b)


def pmhash(b=b""):
    return PMHash(b)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


if __name__ == '__main__':
    from . import Loader

    print("Available loaders:")
    for type_name in sorted(Loader.type_names):
        print(f'  - {type_name}:')
        for prio, class_name, cls in Loader.get_all(type_name):
            print(f'    - {class_name:30s} [{prio}]')



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from abc import ABCMeta, abstractmethod

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ..session import Session

class BasicView(metaclass = ABCMeta):
    @abstractmethod
    def display_error(self, message : str) -> None:
        """Display an error message. No action is expected from user.

        :param message: Message to display
        :type message: str
        """
        pass

    @abstractmethod
    def display_notice(self, message : str) -> None:
        """Display a message. No action is expected from user.

        :param message: Message to display
        :type message: str
        """
        pass

    @abstractmethod
    def display_session_list(self, sessions : list["Session"], ask_to_select_one : bool = False) -> Optional["Session"]:
        """Display a list of Sessions, and optionally ask the user to select one of them.

        :param sessions: List of sessions
        :type sessions: list[Session]
        :param ask_to_select_one: ask user to select a session, defaults to False
        :type ask_to_select_one: bool, optional
        :return: the session selected by user (if asked for it), None otherwise (or if user has cancelled)
        :rtype: Optional[Session]
        """
        pass

    @abstractmethod
    def ask_credentials(self, ask_login : bool = False, ask_password : bool = False, ask_2fa : bool = False) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """Ask user for credentials.

        :param ask_login: Ask for user name, defaults to False
        :type ask_login: bool, optional
        :param ask_password: Ask for the password, defaults to False
        :type ask_password: bool, optional
        :param ask_2fa: Ask for a 2FA code, defaults to False
        :type ask_2fa: bool, optional
        :return: A tuple (login, password, 2fa). Values are None if not asked from the user, or if user cancelled.
        :rtype: tuple[Optional[str], Optional[str], Optional[str]]
        """
        pass




"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import base64
import bcrypt
import os

from proton.session.exceptions import ProtonUnsupportedAuthVersionError


PM_VERSION = 4

SRP_LEN_BYTES = 256
SALT_LEN_BYTES = 10


def bcrypt_b64_encode(s):  # The joy of bcrypt
    bcrypt_base64 = b"./ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" # noqa
    std_base64chars = b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"  # noqa
    s = base64.b64encode(s)
    return s.translate(bytes.maketrans(std_base64chars, bcrypt_base64))


def hash_password_3(hash_class, password, salt, modulus):
    salt = (salt + b"proton")[:16]
    salt = bcrypt_b64_encode(salt)[:22]
    # Manually truncate the password to maximum bcrypt accepted length
    MAX_BCRYPT_PASSWORD_BYTES = 72
    hashed = bcrypt.hashpw(password[:MAX_BCRYPT_PASSWORD_BYTES], b"$2y$10$" + salt)
    return hash_class(hashed + modulus).digest()


def hash_password(hash_class, password, salt, modulus, version):
    if version == 4 or version == 3:
        return hash_password_3(hash_class, password, salt, modulus)

    # If the auth_version is lower then the
    # supported value 3 (which were dropped in 2018). In such a case, the user
    # needs to first login via web so that the auth version can be properly updated.
    #
    # This usually happens on older accounts that haven't been used in a while or
    # account that rarely login via the web client.
    raise ProtonUnsupportedAuthVersionError(
        "Account auth_version is not supported. "
        "Login via webclient for it to  be updated."
    )


def bytes_to_long(s):
    return int.from_bytes(s, 'little')


def long_to_bytes(n, num_bytes):
    return n.to_bytes(num_bytes, 'little')


def get_random(nbytes):
    return bytes_to_long(os.urandom(nbytes))


def get_random_of_length(nbytes):
    offset = (nbytes * 8) - 1
    return get_random(nbytes) | (1 << offset)


def custom_hash(hash_class, *args, **kwargs):
    h = hash_class()
    for s in args:
        if s is not None:
            data = long_to_bytes(s, SRP_LEN_BYTES) if isinstance(s, int) else s
            h.update(data)

    return bytes_to_long(h.digest())



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import List, Any, Optional


class FormField:
    """FormData entry."""
    def __init__(
            self, name: str, value: Any, filename: Optional[str] = None,
            content_type: Optional[str] = None
    ):
        self.name = name
        self.value = value
        self.filename = filename
        self.content_type = content_type


class FormData:
    """Data to be sent as form-encoded data, like an HTML form would."""
    def __init__(self, fields: Optional[List[FormField]] = None):
        self.fields = fields or []

    def add(self, field: FormField):
        """Appends a new field in the form."""
        self.fields.append(field)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from asyncio import transports, TimeoutError
from typing import List
from unittest.mock import Mock
from urllib.parse import urlparse
import json, base64, struct, time, asyncio, random, itertools

from ..exceptions import *
from .base import Transport
from .aiohttp import AiohttpTransport
from .alternativerouting import AlternativeRoutingTransport
from ..api import sync_wrapper


class AutoTransport(Transport):
    # We assume that a given transport fails after that number of seconds
    TRANSPORT_TIMEOUT = 15

    @classmethod
    def _get_priority(cls):
        return 100

    def __init__(self, session, transport_choices: List[Transport] = None, transport_timeout: int = None):
        super().__init__(session)

        self._current_transport = None
        self._transport_choices = transport_choices or [
            (0, AiohttpTransport),
            (5, AlternativeRoutingTransport)
        ]
        self._transport_timeout = transport_timeout or self.TRANSPORT_TIMEOUT

    @property
    def is_available(self) -> bool:
        return self._current_transport is not None

    @property
    def transport_choices(self):
        return self._transport_choices

    @transport_choices.setter
    def transport_choices(self, newvalue):
        self._transport_choices = []

        for timeout, cls in newvalue:
            if not isinstance(cls, Transport):
                raise TypeError("Transports should be a subclass of Transport")
            self._transport_choices.append((timeout, cls))
        self._transport_choices.sort(key=lambda x: x[0])

    async def _ping_via_transport(self, timeout, transport):
        await asyncio.sleep(timeout)
        ping_url = "/tests/ping"
        try:
            result = await asyncio.wait_for(transport.async_api_request(ping_url), self._transport_timeout)
        except TimeoutError as error:
            raise ProtonAPINotReachable(
                f"{type(transport).__name__} transport not available: unable to reach {ping_url}"
            ) from error
        if result != {"Code": 1000}:
            raise ProtonAPINotAvailable(
                f"{type(transport).__name__} transport received unexpected response from {ping_url}:\n"
                f"{result}"
            )
        return transport

    async def find_available_transport(self):
        pending = []
        for timeout, cls in self._transport_choices:
            transport = cls(self._session)
            pending.append(asyncio.create_task(self._ping_via_transport(timeout, transport)))

        results_ok = []
        results_fail = []
        final_timestamp = time.time() + self._transport_timeout
        while len(pending) > 0 and len(results_ok) == 0:
            done, pending = await asyncio.wait(pending, timeout=max(0.1, final_timestamp - time.time()), return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                try:
                    results_ok.append(task.result())
                except (ProtonAPINotAvailable, ProtonAPINotReachable) as e:
                    # That means that we were able to get to the API (wasn't reachable or was mitm'ed)
                    results_fail.append(e)
                except Exception as e:
                    # Unhandled exception, we might want to understand what is going on
                    for task in pending:
                        task.cancel()
                    raise

        for task in pending:
            task.cancel()

        if not results_ok:
            raise ProtonAPINotReachable("No working transports found")

        self._current_transport = results_ok[0]

    async def async_api_request(
        self, endpoint,
        jsondata=None, data=None, additional_headers=None, method=None, params=None,
        return_raw=False
    ):
        tries_left = 3
        while tries_left > 0:
            tries_left -= 1
            if self._current_transport is None:
                await self.find_available_transport()

            try:
                return await asyncio.wait_for(self._current_transport.async_api_request(endpoint, jsondata, data, additional_headers, method, params, return_raw=return_raw), self._transport_timeout)
            except asyncio.TimeoutError:
                # Reset transport
                self._current_transport = None

        raise ProtonAPINotReachable("Timeout accessing the API")  # we should not reach that point except in case of Timeout



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional


class ProtonError(Exception):
    """Base class for Proton API specific exceptions"""
    def __init__(self, message, additional_context=None):
        self.message = message
        self.additional_context = additional_context
        super().__init__(self.message)


class ProtonCryptoError(ProtonError):
    """Exception thrown when something is wrong on the crypto side. 
    In general this has to be handled as being fatal, as something is super-wrong."""


class ProtonUnsupportedAuthVersionError(ProtonCryptoError):
    """When the auth_version returned by the API is lower then what is currently supported.
    This is usually fixed with a login via the webclient."""


class ProtonAPIError(ProtonError):
    """Exception that is raised whenever the API call didn't return a 1000/1001 code.
    Logic for handling these depend on the call (see API doc)
    """

    def __init__(self, http_code, http_headers, json_data):
        self._http_code = http_code
        self._http_headers = http_headers
        self._json_data = json_data

        super().__init__(f'[HTTP/{self.http_code}, {self.body_code}] {self.error}')

    @property
    def http_code(self) -> int:
        """:return: HTTP error code (401, 403, 422...)
        :rtype: int
        """
        return self._http_code

    @property
    def http_headers(self) -> dict:
        """:return: Dictionary of HTTP headers of the error reply
        :rtype: dict
        """
        return self._http_headers

    @property
    def json_data(self) -> dict:
        """:return: JSON data of the error reply
        :rtype: dict
        """
        return self._json_data

    @property
    def body_code(self) -> int:
        """:return: Body error code ("Code" in JSON)
        :rtype: int
        """
        return self._json_data['Code']
    
    @property
    def error(self) -> str:
        """:return: Body error message ("Error" in JSON)
        :rtype: str
        """
        return self._json_data['Error']

    @classmethod
    def from_proton_api_error(cls, e : "ProtonAPIError"):
        """Construct an instance of this class, based on a ProtonAPIError (this allows to downcast to a more specific exception)

        :param e: Initial API exception
        :type e: ProtonAPIError
        :return: An instance of the current class
        :rtype: Any
        """
        return cls(e._http_code, e._http_headers, e._json_data)

class ProtonAPINotReachable(ProtonError):
    """Exception thrown when the transport couldn't reach the API.
    
    One may try using a different transport, or later if the error is transient."""

class ProtonAPINotAvailable(ProtonError):
    """Exception thrown when the API is reachable (i.e. at the TLS level), but doesn't work. 
    
    This is definitive for that transport, it will not work by retrying in the same conditions."""

class ProtonAPIUnexpectedError(ProtonError):
    """Something went wrong, but we don't know how to handle it. Good luck :-)"""

class ProtonAPIAuthenticationNeeded(ProtonAPIError):
    """We tried to call a route that requires authentication, but we don't have it.
    
    This should be solved by calling session.authenticate() with valid credentials"""

class ProtonAPI2FANeeded(ProtonAPIError):
    """We need 2FA authentication, but it's not done yet.
    
    This should be solved by calling session.provide_2fa() with valid 2FA"""

class ProtonAPIMissingScopeError(ProtonAPIError):
    """We don't have a required scope.
    
    This might be because of user rights, but also might require a call to unlock."""

class ProtonAPIHumanVerificationNeeded(ProtonAPIError):
    """Human verification is needed for this API call to succeed."""

    @property
    def HumanVerificationToken(self) -> Optional[str]:
        """Get the Token for human verification"""
        return self.json_data.get('Details', {}).get('HumanVerificationToken', None)

    @property
    def HumanVerificationMethods(self) -> list[str]:
        """Return a list of allowed human verification methods.

        :return: human verification methods
        :rtype: list[str]
        """
        return self.json_data.get('Details', {}).get('Methods', [])



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from os import environ
from proton import session
from proton.session import exceptions
from proton.session.api import Session

import sys

from proton.sso.sso import ProtonSSO

from ..views._base import BasicView

import enum
class ProtonSSOPresenterCredentialLogicState(enum.Enum):
    CALL_BASE_FUNCTION = 0
    NEEDS_AUTHENTICATE = 1
    NEEDS_TWOFA = 2


class ProtonSSOPresenter:
    def __init__(self, view : BasicView, appversion=None, user_agent=None, keyring=None):
        from .sso import ProtonSSO

        self._view = view
        self._session = None
        self._provided_account_name = None
        self._client_secret: str = None

        kwargs_sso = {}
        if appversion is not None:
            kwargs_sso["appversion"] = appversion
        if user_agent is not None:
            kwargs_sso["user_agent"] = user_agent
        if keyring is not None:
            kwargs_sso["keyring_backend_name"] = keyring
        self._sso = ProtonSSO(**kwargs_sso)

    def set_session(self, account_name = None):
        self._provided_account_name = account_name
        if account_name is not None:
            self._session = self._sso.get_session(account_name)
        else:
            self._session = self._sso.get_default_session()

    def set_environment(self, environment):
        self._session.environment = environment

    def set_client_secret(self, client_secret):
        self._client_secret = client_secret

    def CredentialsLogic(base_function):
        import functools

        @functools.wraps(base_function)
        def wrapped_function(self : 'ProtonSSOPresenter', *a, **kw):
            
            from proton.session.exceptions import ProtonAPIAuthenticationNeeded, ProtonAPI2FANeeded, ProtonAPIMissingScopeError
            state = ProtonSSOPresenterCredentialLogicState.CALL_BASE_FUNCTION
            while True:
                try:
                    if state == ProtonSSOPresenterCredentialLogicState.CALL_BASE_FUNCTION:
                        return base_function(self, *a, **kw)
                    elif state == ProtonSSOPresenterCredentialLogicState.NEEDS_AUTHENTICATE:
                        account_name, password, twofa = self._view.ask_credentials(self._provided_account_name is None, True, False)
                        if account_name is None:
                            account_name = self._provided_account_name
                        if password is None:
                            break
                        ret = self._session.authenticate(account_name, password, client_secret=self._client_secret),
                        if ret:
                            if self._session.needs_twofa:
                                state = ProtonSSOPresenterCredentialLogicState.NEEDS_TWOFA
                            else:
                                state = ProtonSSOPresenterCredentialLogicState.CALL_BASE_FUNCTION
                        else:
                            self._view.display_error("Invalid credentials!")
                            # Remain in NEEDS_AUTHENTICATE state
                    elif state == ProtonSSOPresenterCredentialLogicState.NEEDS_TWOFA:
                        if self._session.supports_fido2:
                            self._view.display_notice("This session supports 2FA via FIDO2 but proton.sso is a simple tool not supporting it, sorry => 2FA is likely to fail if it's the only available 2FA...")
                        account_name, password, twofa = self._view.ask_credentials(False, False, True)
                        if twofa is None:
                            break
                        ret = self._session.validate_2fa_code(twofa)
                        if ret:
                            state = ProtonSSOPresenterCredentialLogicState.CALL_BASE_FUNCTION
                        else:
                            self._view.display_error("Invalid 2FA code!")

                except ProtonAPIAuthenticationNeeded:
                    state = ProtonSSOPresenterCredentialLogicState.NEEDS_AUTHENTICATE
 
                except ProtonAPI2FANeeded:
                    state = ProtonSSOPresenterCredentialLogicState.NEEDS_TWOFA


        
        return wrapped_function


    @CredentialsLogic
    def login(self):
        # This will force a login call if needed
        self._session.api_request('/users')

    def logout(self):
        self._session.logout()

    def unlock(self):
        self._session.fetch_user_key()

    def lock(self):
        self._session.lock()

    def set_default(self):
        account_name = self._session.AccountName
        if account_name is not None:
            self._sso.set_default_account(account_name)

    def list(self):
        sessions = [self._sso.get_session(s) for s in self._sso.sessions]
        sessions = [s for s in sessions if s.AccountName is not None]
        self._view.display_session_list(sessions)


def main():
    import argparse

    parser = argparse.ArgumentParser('proton-sso', description="Tool to manage user SSO sessions")
    parser.add_argument('--appversion', help="App version")
    parser.add_argument('--user-agent', help="User Agent")
    parser.add_argument('--keyring', help="Keyring (if not set, will select default keyring)")
    subparsers = parser.add_subparsers(help='action', dest='action', required=True)

    parser_login = subparsers.add_parser('login', help='Sign into an account')
    parser_login.add_argument('--unlock', action='store_true', help="Unlock and store user keys")
    parser_login.add_argument('--set-default', action='store_true', help="Set this account as default")
    parser_login.add_argument('--env', type=str, help="Environment to use")
    parser_login.add_argument('--client-secret', type=str, help="Some API require a client secret")
    parser_login.add_argument('account', type=str, help="Proton account")

    parser_logout = subparsers.add_parser('logout', help='Sign out of an account')
    parser_logout.add_argument('account', type=str, help="Proton account (default session if omitted)", nargs="?")

    parser_lock = subparsers.add_parser('lock', help='Lock a session and erased stored user keys')
    parser_lock.add_argument('account', type=str, help="Proton account (default session if omitted)", nargs="?")

    parser_unlock = subparsers.add_parser('unlock', help='Unlock a session and store user keys')
    parser_unlock.add_argument('account', type=str, help="Proton account (default session if omitted)", nargs="?")

    parser_unlock = subparsers.add_parser('set-default', help='Sets the account as default')
    parser_unlock.add_argument('account', type=str, help="Proton account")
    
    parser_list = subparsers.add_parser('list', help='List the currently logged-in account')
    args = parser.parse_args()

    from proton.loader import Loader

    view = Loader.get('basicview')()
    presenter = ProtonSSOPresenter(view, appversion=args.appversion, user_agent=args.user_agent, keyring=args.keyring)

    # All action except list require an active account
    if args.action != 'list':
        presenter.set_session(args.account)

    if args.action == 'login':
        if args.env is not None:
            presenter.set_environment(args.env)
        if args.client_secret is not None:
            presenter.set_client_secret(args.client_secret)
        presenter.login()
        if args.unlock:
            presenter.unlock()
        else:
            presenter.lock()
        if args.set_default:
            presenter.set_default()

    elif args.action == 'logout':
        presenter.logout()

    elif args.action == 'lock':
        presenter.lock()

    elif args.action == 'unlock':
        presenter.unlock()
    elif args.action == 'list':
        presenter.list()
    elif args.action == 'set-default':
        presenter.set_default()
    else:
        raise NotImplementedError(f"Action {args.action} is not yet implemented")


if __name__ == '__main__':
    main()



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
# N    A large safe prime (N = 2q+1, where q is prime)
#      All arithmetic is done modulo N.
# g    A generator modulo N
# k    Multiplier parameter (k = H(N, g) in SRP-6a, k = 3 for legacy SRP-6)
# s    User's salt
# I    Username
# p    Cleartext Password
# H()  One-way hash function
# ^    (Modular) Exponentiation
# u    Random scrambling parameter
# a,b  Secret ephemeral values
# A,B  Public ephemeral values
# x    Private key (derived from p and s)
# v    Password verifier
from .pmhash import pmhash
from .util import (PM_VERSION, SRP_LEN_BYTES, SALT_LEN_BYTES, bytes_to_long, custom_hash,
                   get_random_of_length, hash_password, long_to_bytes)


def get_ng(n_bin, g_hex):
    return bytes_to_long(n_bin), int(g_hex, 16)


def hash_k(hash_class, g, modulus, width):
    h = hash_class()
    h.update(g.to_bytes(width, 'little'))
    h.update(modulus.to_bytes(width, 'little'))
    return bytes_to_long(h.digest())


def calculate_x(hash_class, salt, password, modulus, version):
    exp = hash_password(
        hash_class, password, salt, long_to_bytes(modulus, SRP_LEN_BYTES), version
    )
    return bytes_to_long(exp)


def calculate_client_proof(hash_class, A, B, K):
    h = hash_class()
    h.update(long_to_bytes(A, SRP_LEN_BYTES))
    h.update(long_to_bytes(B, SRP_LEN_BYTES))
    h.update(K)
    return h.digest()


def calculate_server_proof(hash_class, A, M, K):
    h = hash_class()
    h.update(long_to_bytes(A, SRP_LEN_BYTES))
    h.update(M)
    h.update(K)
    return h.digest()


class User(object):
    def __init__(self, password, n_bin, g_hex=b"2", bytes_a=None, bytes_A=None): # noqa
        if bytes_a and len(bytes_a) != 32:
            raise ValueError("32 bytes required for bytes_a")

        if not isinstance(password, str) or len(password) == 0:
            raise ValueError("Invalid password")

        self.N, self.g = get_ng(n_bin, g_hex)
        self.hash_class = pmhash
        self.k = hash_k(
            self.hash_class, self.g,
            self.N, SRP_LEN_BYTES
        )

        self.p = password.encode()
        if bytes_a:
            self.a = bytes_to_long(bytes_a)
        else:
            self.a = get_random_of_length(32)
        if bytes_A:
            self.A = bytes_to_long(bytes_A)
        else:
            self.A = pow(self.g, self.a, self.N)
        self.v = None
        self.M = None
        self.K = None
        self.expected_server_proof = None
        self._authenticated = False
        self.bytes_s = None
        self.S = None
        self.B = None
        self.u = None
        self.x = None

    def authenticated(self):
        return self._authenticated

    def get_ephemeral_secret(self):
        return long_to_bytes(self.a, SRP_LEN_BYTES)

    def get_session_key(self):
        return self.K if self._authenticated else None

    def get_challenge(self):
        return long_to_bytes(self.A, SRP_LEN_BYTES)

    # Returns M or None if SRP-6a safety check is violated
    def process_challenge(
        self, bytes_s, bytes_server_challenge, version=PM_VERSION
    ):
        self.bytes_s = bytes_s
        self.B = bytes_to_long(bytes_server_challenge)

        # SRP-6a safety check
        if (self.B % self.N) == 0:
            return None

        self.u = custom_hash(self.hash_class, self.A, self.B)

        # SRP-6a safety check
        if self.u == 0:
            return None

        self.x = calculate_x(self.hash_class, self.bytes_s, self.p, self.N, version)

        self.v = pow(self.g, self.x, self.N)

        self.S = pow(
            (self.B - self.k * self.v), (self.a + self.u * self.x), self.N
        )

        self.K = long_to_bytes(self.S, SRP_LEN_BYTES)
        self.M = calculate_client_proof(self.hash_class, self.A, self.B, self.K) # noqa
        self.expected_server_proof = calculate_server_proof(
            self.hash_class, self.A, self.M, self.K
        )

        return self.M

    def verify_session(self, server_proof):
        if self.expected_server_proof == server_proof:
            self._authenticated = True

    def compute_v(self, bytes_s=None, version=PM_VERSION):
        self.bytes_s = long_to_bytes(get_random_of_length(SALT_LEN_BYTES), SALT_LEN_BYTES) if bytes_s is None else bytes_s # noqa
        self.x = calculate_x(self.hash_class, self.bytes_s, self.p, self.N, version)

        return self.bytes_s, long_to_bytes(pow(self.g, self.x, self.N), SRP_LEN_BYTES) # noqa



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import ipaddress
import logging
import re
import struct
import typing
import random


class DNSResponseError(Exception):
    pass


class DNSParsingException(DNSResponseError):
    pass


class DNSParser:
    """Parse response from any DNS resolvers"""

    STRUCT_REPLY_COUNTS = struct.Struct('>HHHH')
    STRUCT_REC_FORMAT = struct.Struct('>HHIH')

    _MINIMUM_RECORD_LENGTH = 12  # => Transaction ID/Flags/#Questions/#Answers/#AuthorRRs/#AdditRRs

    # Regular expression used in _is_valid_hostname.
    # Each hostname segment (the strings in between the period characters) is only valid if:
    #  - it has a minimum of 1 character and maximum of 63,
    #  - it only contains alphanumeric characters or the hyphen but
    #  - it does not start or end with a hyphen.
    _VALID_HOSTNAME_SEGMENT = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)

    # type definitions
    IPvxAddress = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
    ParsedData = typing.Union[str, IPvxAddress]

    @classmethod
    def parse(cls, reply_data) -> typing.Optional[typing.List[typing.Tuple[int, ParsedData]]]:
        """ parse DNS reply and returns list of : TTL, address(IP or CNAME)"""

        if len(reply_data) < cls._MINIMUM_RECORD_LENGTH:
            raise DNSParsingException(f"(truncated reply)")

        # Match reply code (0x0 = OK)
        dns_rcode = reply_data[3] & 0xf

        # ensure we have something to parse
        if dns_rcode == 0x3:
            #NXDOMAIN, this is fatal
            raise DNSResponseError("No alternative routing exists for this environment (NXDOMAIN)")
        elif dns_rcode != 0x0:
            raise DNSResponseError(f"DNS response error code: {dns_rcode}")

        # Get counts
        offset = 4
        dns_qdcount, dns_ancount, dns_nscount, dns_arcount = cls.STRUCT_REPLY_COUNTS.unpack_from(reply_data[offset:])
        offset += cls.STRUCT_REPLY_COUNTS.size
        # skip questions
        for dns_qd_idx in range(dns_qdcount):
            length, data = cls._get_name(reply_data, offset)
            # We ignore QTYPE/QCLASS
            offset += length + 4

        # tuples (TTL, data)
        answers = []
        # answers
        for dns_an_idx in range(dns_ancount):
            length, data = cls._get_name(reply_data, offset)

            offset += length
            try:
                rec_type, rec_class, rec_ttl, rec_dlen = cls.STRUCT_REC_FORMAT.unpack_from(reply_data[offset:])
            except struct.error:
                raise DNSParsingException(f"(truncated record headers)")
            offset += cls.STRUCT_REC_FORMAT.size

            record = reply_data[offset:offset + rec_dlen]
            if offset + rec_dlen > len(reply_data):
                raise DNSParsingException(f"(truncated reply while parsing record)")
            offset += rec_dlen

            if rec_type == 0x10 and rec_class == 0x01:  # IN TXT
                if record[0] != rec_dlen - 1:
                    raise DNSParsingException(f"(length of TXT record doesn't match REC_DLEN)")
                if record[0] != len(record) - 1:
                    raise DNSParsingException(f"(length of TXT record doesn't actual record data)")
                try:
                    hostname = record[1:].decode('ascii')

                    # Only hostnames with a valid format are accepted in TXT records.
                    # This is to avoid possible security exploits where the TXT record contains
                    # a full URL (hostname + path), which e.g. could trigger SSO redirects.
                    if not cls._is_valid_hostname(hostname):
                        raise DNSParsingException(f"Invalid hostname in TXT record: {hostname}")

                    answers.append((int(rec_ttl), hostname))
                except UnicodeDecodeError:
                    raise DNSParsingException(f"(UnicodeDecodeError in TXT record)")
            elif rec_type == 0x01 and rec_class == 0x01:  # IN A
                if len(record) != 4:
                    raise DNSParsingException(f"(length of A record doesn't match)")

                try:
                    ipv4_address = ipaddress.IPv4Address(record)
                except ValueError as exc:
                    raise DNSParsingException(f"Invalid IP address in A record: {record}") from exc

                answers.append((int(rec_ttl), ipv4_address))
            else:
                logging.debug(f"record type currently not supported: {rec_type}... skip")

        return answers

    @staticmethod
    def _get_name(buffer: bytes, offset=0):
        # Length that we've parsed (for that specific record)
        parsed_length = 0
        # Have we jumped to somewhere else?
        has_jumped = False
        # Parts we've seen until now
        parts = []
        # While we're in the buffer, and we are not on a null byte (terminator)
        while offset < len(buffer) and buffer[offset] != 0:
            # Read the length of the part, in one byte
            length = buffer[offset]
            # If the length starts with two one bytes, then it's a pointer
            if length & 0b1100_0000 == 0b1100_0000:
                offset = ((buffer[offset] & 0b0011_1111) << 8) + buffer[offset + 1]
                # Pointers have length 2
                if not has_jumped:
                    parsed_length += 2
                # We're not any more in the current record, stop counting
                has_jumped = True
            else:
                # Real entry
                # Add the part
                if offset + 1 + length > len(buffer):
                    raise DNSParsingException(f"DNS resolution failed (non-parsable value)")
                parts.append(buffer[offset + 1:offset + 1 + length])
                # Add length, and the length byte
                if not has_jumped:
                    parsed_length += length + 1
                offset += length + 1

        # This is for the 0-byte that terminates a name
        if not has_jumped:
            parsed_length += 1

        return parsed_length, parts

    @classmethod
    def _build_simple_query(cls, domain: bytes, qtype: int, qclass: int):
        """internal utility to build the simplest DNS request we need"""
        id: bytes = struct.pack('!H', random.randint(0, 65535))  # nosemgrep: gitlab.bandit.B311 # nosec B311 # noqa: E501 # pylint: disable=line-too-long
        qtype: bytes = struct.pack('!H', qtype)
        qclass: bytes = struct.pack('!H', qclass)

        # it's a query with a single question, no AN, no RR, no AR
        return id + b"\x01\x20\x00\x01\x00\x00\x00\x00\x00\x00" + domain + qtype + qclass

    @classmethod
    def build_query(cls, fqdn: typing.Union[str, bytes], qtype, qclass):
        """Build a very simple dns request that is just a query with no AN, no RR, no AR, and a single question"""

        if type(fqdn) == str:
            domain = b''.join([bytes([len(el)]) + el.encode('ascii') for el in fqdn.split('.')]) + b'\x00'
        elif type(fqdn) == bytes:
            domain = b''.join([bytes([len(el)]) + el for el in fqdn.split(b'.')]) + b'\x00'
        else:
            raise TypeError("fqdn can only be str or bytes")

        query = cls._build_simple_query(domain, qtype, qclass)
        return query

    @classmethod
    def _is_valid_hostname(cls, hostname):
        if len(hostname) > 255:
            return False

        # Strip exactly one dot from the right, if present.
        if hostname[-1] == ".":
            hostname = hostname[:-1]

        # The hostname is valid if all its segments are valid.
        return all(cls._VALID_HOSTNAME_SEGMENT.match(segment) for segment in hostname.split("."))



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import *
import logging

from .exceptions import ProtonCryptoError, ProtonAPIError, ProtonAPIAuthenticationNeeded, ProtonAPI2FANeeded, ProtonAPIMissingScopeError, ProtonAPIHumanVerificationNeeded
from .srp import User as PmsrpUser
from .environments import Environment
from ..loader import Loader

import asyncio
import base64
import functools
import random

from ..utils import ExecutionEnvironment

logger = logging.getLogger(__name__)

SRP_MODULUS_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----

xjMEXAHLgxYJKwYBBAHaRw8BAQdAFurWXXwjTemqjD7CXjXVyKf0of7n9Ctm
L8v9enkzggHNEnByb3RvbkBzcnAubW9kdWx1c8J3BBAWCgApBQJcAcuDBgsJ
BwgDAgkQNQWFxOlRjyYEFQgKAgMWAgECGQECGwMCHgEAAPGRAP9sauJsW12U
MnTQUZpsbJb53d0Wv55mZIIiJL2XulpWPQD/V6NglBd96lZKBmInSXX/kXat
Sv+y0io+LR8i2+jV+AbOOARcAcuDEgorBgEEAZdVAQUBAQdAeJHUz1c9+KfE
kSIgcBRE3WuXC4oj5a2/U3oASExGDW4DAQgHwmEEGBYIABMFAlwBy4MJEDUF
hcTpUY8mAhsMAAD/XQD8DxNI6E78meodQI+wLsrKLeHn32iLvUqJbVDhfWSU
WO4BAMcm1u02t4VKw++ttECPt+HUgPUq5pqQWe5Q2cW4TMsE
=Y4Mw
-----END PGP PUBLIC KEY BLOCK-----"""

SRP_MODULUS_KEY_FINGERPRINT = "248097092b458509c508dac0350585c4e9518f26"  # nosemgrep: gitleak-detect-ignore,generic.secrets.gitleaks.generic-api-key.generic-api-key # pylint: disable=line-too-long


@dataclass
class Fido2AssertionParameters:
    """
    Parameters needed to create a FIDO2 assertion for 2FA.
    These can be obtained from :attr:`Session.supports_fido2`.
    """
    challenge: bytes
    rp_id: str
    allow_credentials: list[dict]
    user_verification: str


@dataclass
class Fido2Assertion:
    """
    FIDO2 assertion obtained from the security key.
    These are passed to :meth:`Session.async_validate_2fa_fido2`.
    """
    client_data: bytes
    authenticator_data: bytes
    signature: bytes
    credential_id: bytes


def sync_wrapper(f):
    @functools.wraps(f)
    def wrapped_f(*a, **kw):
        try:
            loop = asyncio.get_running_loop()
            newloop = False
        except RuntimeError:
            newloop = True

        if not newloop:
            raise RuntimeError("It's forbidden to call sync_wrapped functions from an async one, please await directly the async one")
        
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(f(*a, **kw))
        finally:
            loop.close()
    wrapped_f.__doc__ = f"Synchronous wrapper for :meth:`{f.__name__}`" + "\n\n" + (wrapped_f.__doc__ or "")
    return wrapped_f

class Session:
    def __init__(self, appversion : str = "Other", user_agent:str="None"):
        """Get a session towards the Proton API.

        :param appversion: version for the new Session object, defaults to ``"Other"``
        :type appversion: str, optional
        :param user_agent: user agent to use, defaults to ``"None"``. It should be of the following syntax:

          * Linux based -> ``ClientName/client.version (Linux; Distro/distro_version)``
          * Non-linux based -> ``ClientName/client.version (OS)``

        :type user_agent: str, optional
        """
        self.__appversion = appversion
        self.__user_agent = user_agent

        self.__UID = None
        self.__AccessToken = None
        self.__RefreshToken = None
        self.__Scopes = None

        self.__AccountName = None

        #Extra data that we want to persist (used if we load a session from a subclass)
        self.__extrastate = {}

        # Temporary storage for 2FA object
        self.__2FA = None

        #Refresh revision (incremented each time a refresh is done)
        #This allows knowing if a refresh should be done or if it is already in progress
        self.__refresh_revision = 0

        #Lazy initialized by modulus decryption
        self.__gnupg_for_modulus = None

        #Lazy initialized by api request
        self.__transport = None
        self.__transport_factory = None

        self.transport_factory = None

        #Lazy initialized by request lock/unlock
        self.__can_run_requests = None

        #Lazy initialized by environment:
        self.__environment = None

        self.__persistence_observers = []


    async def async_api_request(self, endpoint,
        jsondata=None, data=None, additional_headers=None,
        method=None, params=None, no_condition_check=False,
        return_raw=False):
        """Do an API request.

        This call can return any of the exceptions defined in :mod:`proton.session.exceptions`.

        :param endpoint: API endpoint
        :type endpoint: str
        :param jsondata: JSON serializable dict to send as request data
        :type jsondata: dict
        :param data: data to be sent as either `multipart/form-data` or
            `application/x-www-form-urlencoded`. `multipart/form-data` is used
            when required, for example if data includes fields with a file-like
            value (i.e. is an instance of io.IOBase).
        :type data: FormData
        :param additional_headers: additional headers to send
        :type additional_headers: dict
        :param method: HTTP method (get|post|put|delete|patch)
        :type method: str
        :param params: URL parameters to append to the URL. If a dictionary or
            list of tuples ``[(key, value)]`` is provided, form-encoding will
            take place.
        :type params: str, dict or iterable
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param return_raw: If set to true, the function will return a :class:`RawResponse` object
            instead of a dict, this class contains the header and status information along with the
            json response.
        :type return_raw: bool, optional

        :return: Deserialized JSON reply or a RawResponse object if return_raw is True
        :rtype: dict | RawResponse
        """

        # We might need to loop
        attempts = 3
        stored_exception = None  # in case of too many attempts, raise instead of returning None
        while attempts > 0:
            attempts -= 1
            try:
                refresh_revision_at_start = self.__refresh_revision
                return await self.__async_api_request_internal(endpoint, jsondata, data, additional_headers, method, params, no_condition_check,
                                                               return_raw=return_raw)
            except ProtonAPIError as e:
                stored_exception = e
                # We have a missing scope.
                if e.http_code == 403:
                    # If we need a 2FA authentication, then ask for it by sending a specific exception.
                    if self.needs_twofa:
                        raise ProtonAPI2FANeeded.from_proton_api_error(e)
                    else:
                        # Otherwise, just throw the 403
                        raise ProtonAPIMissingScopeError.from_proton_api_error(e)
                #401: token expired
                elif e.http_code == 401:
                    #If we can refresh, than do it and retry
                    if await self.async_refresh(only_when_refresh_revision_is=refresh_revision_at_start, no_condition_check=no_condition_check):
                        continue
                    #Else, fail :-(
                    else:
                        raise ProtonAPIAuthenticationNeeded.from_proton_api_error(e)
                #422 + 9001: Human verification needed
                elif e.http_code == 422 and e.body_code == 9001:
                    raise ProtonAPIHumanVerificationNeeded.from_proton_api_error(e)
                #Invalid human verification token
                elif e.body_code == 12087:
                    raise ProtonAPIHumanVerificationNeeded.from_proton_api_error(e)
                #These are codes which require and immediate retry
                elif e.http_code in (408, 502):
                    continue
                #These not, let's retry more gracefully
                elif e.http_code in (429, 503):
                    await self.__sleep_for_exception(e)
                    continue
                #Something else, throw
                raise
        raise stored_exception  # if we have reached that point without returning any value, an exception should have been stored

    async def async_authenticate(self, username: str, password: str, client_secret: str = None, no_condition_check: bool = False, additional_headers=None) -> bool:
        """Authenticate against Proton API

        :param username: Proton account username
        :type username: str
        :param password: Proton account password
        :type password: str
        :param client_secret: Client Secret for SRP
        :type client_secret: str, optional
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param additional_headers: additional headers to send
        :type additional_headers: dict
        :return: True if authentication succeeded, False otherwise.
        :rtype: bool
        """
        self._requests_lock(no_condition_check)

        await self.async_logout(no_condition_check=True)

        try:
            req_data = {"Username": username}
            if client_secret is not None:
                req_data["ClientSecret"] = client_secret
            info_response = await self.__async_api_request_internal("/auth/info", req_data,
                                                                    no_condition_check=True,
                                                                    additional_headers=additional_headers)

            modulus = self._verify_modulus(info_response['Modulus'])
            server_challenge = base64.b64decode(info_response["ServerEphemeral"])
            salt = base64.b64decode(info_response["Salt"])
            version = info_response["Version"]

            usr = PmsrpUser(password, modulus)
            client_challenge = usr.get_challenge()
            client_proof = usr.process_challenge(salt, server_challenge, version)

            if client_proof is None:
                raise ProtonCryptoError('Invalid challenge')

            # Send response
            payload = {
                "Username": username,
                "ClientEphemeral": base64.b64encode(client_challenge).decode(
                    'utf8'
                ),
                "ClientProof": base64.b64encode(client_proof).decode('utf8'),
                "SRPSession": info_response["SRPSession"],
            }
            if client_secret is not None:
                payload["ClientSecret"] = client_secret
            try:
                auth_response = await self.__async_api_request_internal("/auth", payload, no_condition_check=True,
                                                                        additional_headers=additional_headers)
            except ProtonAPIError as e:
                if e.body_code == 8002:
                    return False
                raise

            if "ServerProof" not in auth_response:
                return False

            usr.verify_session(base64.b64decode(auth_response["ServerProof"]))
            if not usr.authenticated():
                raise ProtonCryptoError('Invalid server proof')

            self.__UID = auth_response['UID']
            self.__AccessToken = auth_response['AccessToken']
            self.__RefreshToken = auth_response['RefreshToken']
            self.__Scopes = auth_response["Scopes"]
            self.__AccountName = username

            if '2FA' in auth_response:
                self.__2FA = auth_response['2FA']
            else:
                self.__2FA = None

            return True
        finally:
            self._requests_unlock(no_condition_check)

    async def async_provide_2fa(self, code : str, no_condition_check=False, additional_headers=None) -> bool:
        """Provide Two Factor Authentication Code to the API.

        This method is deprecated, please use :meth:`async_validate_2fa_code`
        or :meth:`async_validate_2fa_fido2` instead.
        
        :param code: 2FA code
        :type code: str
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :return: True if 2FA succeeded, False otherwise.
        :rtype: bool
        :raises ProtonAPIAuthenticationNeeded: if 2FA failed, and the session was reset by the API backend (this is normally the case)
        """

        logger.warning("async_provide_2fa is deprecated, please use async_validate_2fa_code or async_validate_2fa_fido2 instead")
        
        return await self.async_validate_2fa_code(code, no_condition_check,
                                                  additional_headers)

    async def _async_validate_2fa(self, answer: dict, no_condition_check=False, additional_headers=None) -> bool:
        """
        Internal function to validate 2FA, either via code or FIDO2.

        Do not use this method directly, use :meth:`async_validate_2fa_code`
        or :meth:`async_validate_2fa_fido2` instead.

        :param answer: dict containing either the TwoFactorCode or FIDO2 information
        :type answer: dict
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param additional_headers: additional headers to send
        :type additional_headers: dict, optional
        :return: True if 2FA succeeded, False otherwise.
        """
        self._requests_lock(no_condition_check)
        try:
            ret = await self.__async_api_request_internal(
                '/auth/2fa', answer, no_condition_check=True,
                additional_headers=additional_headers)

            self.__Scopes = ret['Scopes']
            if ret.get('Code') == 1000:
                self.__2FA = None
                return True
            
            return False
        except ProtonAPIError as e:
            if e.body_code == 8002:
                # 2FA jail, we need to start over (beware, we might hit login jails too)
                #Needs re-login
                self._clear_local_data()
                raise ProtonAPIAuthenticationNeeded.from_proton_api_error(e)
            if e.http_code == 401:
                return False
            raise
        finally:
            self._requests_unlock(no_condition_check)

    async def async_validate_2fa_code(self,
                                      code: str,
                                      no_condition_check=False,
                                      additional_headers=None) -> bool:
        """
        Validate a 2FA code against the API.
        :param code: 2FA code
        :type code: str
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param additional_headers: additional headers to send
        :type additional_headers: dict, optional
        :return: True if 2FA succeeded, False otherwise.
        :rtype: bool
        """

        return await self._async_validate_2fa(
            {"TwoFactorCode": code},
            no_condition_check,
            additional_headers)

    async def async_validate_2fa_fido2(self,
                                       assertion: Fido2Assertion,
                                       no_condition_check=False,
                                       additional_headers=None) -> bool:
        """
        Validate a FIDO2 assertion against the API.
        :param assertion: FIDO2 assertion obtained from the security key
        :type assertion: Fido2Assertion
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param additional_headers: additional headers to send
        :type additional_headers: dict, optional
        :return: True if 2FA succeeded, False otherwise.
        :rtype: bool
        """

        fido2 = self.__2FA["FIDO2"]

        def to_b64(x): return base64.b64encode(x).decode('ascii')

        answer = {
            "AuthenticationOptions": fido2["AuthenticationOptions"],
            "ClientData": to_b64(assertion.client_data),
            "AuthenticatorData": to_b64(assertion.authenticator_data),
            "Signature": to_b64(assertion.signature),
            "CredentialID": list(assertion.credential_id),
        }
        return await self._async_validate_2fa(
            {"FIDO2": answer},
            no_condition_check,
            additional_headers
        )

    async def async_refresh(self, only_when_refresh_revision_is=None, no_condition_check=False, additional_headers=None):
        """Refresh tokens.

        Refresh AccessToken with a valid RefreshToken.
        If the RefreshToken is invalid then the user will have to
        re-authenticate.

        :return: True if refresh succeeded, False otherwise (doesn't throw an exception)
        :rtype: bool
        """

        #If we have the correct revision, and it doesn't match, then just exit
        if only_when_refresh_revision_is is not None and only_when_refresh_revision_is != self.__refresh_revision:
            # If we have the wrong revision, then this indicates that we have two refresh running in parallel.
            # Thanksfully, we can simply wait for the other to complete and return successfully.
            await self._requests_wait(no_condition_check)
            return True

        self._requests_lock(no_condition_check)

        #Increment the refresh revision counter, so we don't refresh multiple times
        self.__refresh_revision += 1

        attempts = 3

        try:
            while attempts > 0:
                attempts -= 1
                try:
                    refresh_response = await self.__async_api_request_internal('/auth/refresh', {
                        "ResponseType": "token",
                        "GrantType": "refresh_token",
                        "RefreshToken": self.__RefreshToken,
                        "RedirectURI": "http://protonmail.ch"
                    }, no_condition_check=True, additional_headers=additional_headers)
                    self.__AccessToken = refresh_response["AccessToken"]
                    self.__RefreshToken = refresh_response["RefreshToken"]
                    self.__Scopes = refresh_response["Scopes"]
                    return True

                except ProtonAPIError as e:
                    #https://confluence.protontech.ch/display/API/Authentication%2C+sessions%2C+and+tokens#Authentication,sessions,andtokens-RefreshingSessions
                    if e.http_code == 409:
                        #409 Conflict - Indicates a race condition on the DB, and the request should be performed again
                        continue
                    #We're probably jailed, just retry later
                    elif e.http_code in (429, 503):
                        await self.__sleep_for_exception(e)
                        continue
                    elif e.http_code in (400, 422):
                        #Needs re-login
                        self._clear_local_data()
                        return False
                    return False
        finally:
            self._requests_unlock(no_condition_check)



    async def async_logout(self, no_condition_check=False, additional_headers=None):
        """Logout from API.
        
        :return: True if logout was successful (or nothing was done)
        :rtype: bool
        """

        self._requests_lock(no_condition_check)
        previous_account_name = self.AccountName
        try:
            # No-op if not authenticated (but we do this inside the lock, so data is persisted nevertheless)
            if not self.authenticated:
                self._clear_local_data()
                return True

            ret = await self.__async_api_request_internal('/auth', method='DELETE', no_condition_check=True,
                                                          additional_headers=additional_headers)
            # Erase any information we have about the session
            self._clear_local_data()
            return True
        except ProtonAPIError as e:
            # If we get a 401, then we should erase data (session doesn't exist on the server), and we're fine
            if e.http_code == 401:
                self._clear_local_data()
                return True
            # We don't know what is going on, throw
            raise

        finally:
            self._requests_unlock(no_condition_check, previous_account_name)

    async def async_lock(self, no_condition_check=False, additional_headers=None):
        """ Lock the current user (remove PASSWORD and LOCKED scopes)"""

        self._requests_lock(no_condition_check)
        try:
            ret = await self.__async_api_request_internal('/users/lock', method='PUT', no_condition_check=True,
                                                          additional_headers=additional_headers)
            ret = await self.__async_api_request_internal('/auth/scopes', no_condition_check=True,
                                                          additional_headers=additional_headers)
            self.__Scopes = ret['Scopes']
            return True
        finally:
            self._requests_unlock(no_condition_check)
        #FIXME: clear user keys

    #FIXME: implement unlock

    async def async_human_verif_request_code(self, address=None, phone=None, additional_headers=None):
        """Request a verification code. Either address (email address) or phone (phone number) should be specified."""
        # We use email validation by default if both are provided,
        # but it's not super clean if the dev doesn't know about it
        assert address is not None ^ phone is not None  # nosec B101 # nosemgrep: gitlab.bandit.B101

        if address is not None:
            data = {'Type': 'email', 'Destination': {'Address': address}}
        elif phone is not None:
            data = {'Type': 'sms', 'Destination': {'Phone': phone}}
        
        return await self.async_api_request('/users/code', data, additional_headers=additional_headers).get('Code', 0) == 1000

    async def async_human_verif_provide_token(self, method, token):
        pass


    async def async_fork(self, child_client_id: str, payload: str=None, independent: int=0, user_code: int=None, no_condition_check=False) -> str:
        """Try to fork the current session with parameters and return the fork selector if successful.

        :param child_client_id: The clientID that the client creating the child session will use.
        :type child_client_id: str
        :param payload: Will be downloaded by the child client, and can contain encrypted sensitive information.
        :type payload: str, optional
        :param independent: set to 1 if the newly forked session has to be independent, else 0.
        :type independent: int, optional
        :param user_code: If provided, the selector will not be randomly chosen, but rather the one already returned will be used.
        :type user_code: int, optional
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional

        :return selector: Fork selector to be used by method `async_import_fork()`
        :rtype: str
        """
        data = {'ChildClientID' : child_client_id, 'Independent': independent}
        if payload is not None:
            data['Payload'] = payload
        if user_code is not None:
            data['UserCode'] = user_code
        ret = await self.async_api_request('/auth/v4/sessions/forks', data, method='POST', no_condition_check=no_condition_check)
        return ret['Selector']


    async def async_import_fork(self, selector: str, no_condition_check=False) -> str:
        """Try to import the session fork specified by the selector.

        :param selector: Obtained via a successful call to `async_fork()`
        :type selector: str
        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional

        :return payload: The payload sent by the parent session when initiating the fork.
        :rtype: str
        """
        self._requests_lock(no_condition_check)
        try:
            ret = await self.async_api_request(f"/auth/v4/sessions/forks/{selector}", method='GET', no_condition_check=True)
            self.__UID = ret['UID']
            self.__RefreshToken = ret['RefreshToken']
            self.__AccessToken = ret['AccessToken']
            self.__Scopes = ret['Scopes']
            return ret['Payload']
        finally:
            self._requests_unlock(no_condition_check)
        


    # Wrappers to provide non-asyncio API
    api_request = sync_wrapper(async_api_request)
    authenticate = sync_wrapper(async_authenticate)
    provide_2fa = sync_wrapper(async_provide_2fa)
    logout = sync_wrapper(async_logout)
    refresh = sync_wrapper(async_refresh)
    lock = sync_wrapper(async_lock)
    human_verif_request_code = sync_wrapper(async_human_verif_request_code)
    human_verif_provide_token = sync_wrapper(async_human_verif_provide_token)
    fork = sync_wrapper(async_fork)
    import_fork = sync_wrapper(async_import_fork)
    validate_2fa_code = sync_wrapper(async_validate_2fa_code)
    validate_2fa_fido2 = sync_wrapper(async_validate_2fa_fido2)

    def register_persistence_observer(self, observer: object):
        """Register an observer that will be notified of any persistent state change of the session

        :param observer: Observer to register. It has to provide the following interface (see :class:`proton.sso.ProtonSSO` for an actual implementation):

          * ``_acquire_session_lock(account_name : str, session_data : dict)``
          * ``_release_session_lock(account_name : str, new_session_data : dict)``

        :type observer: object
        """
        self.__persistence_observers.append(observer)

    def _clear_local_data(self) -> None:
        """Clear locally cache data for logout (or equivalently, when the session is "lost")."""
        self.__UID = None
        self.__AccessToken = None
        self.__RefreshToken = None
        self.__Scopes = None
        self.__2FA = None
        self.__extrastate = {}

    @property
    def transport_factory(self):
        """Set/read the factory used for transports (i.e. how to reach the API).

        If the property is set to a class, it will be wrapped in a factory.

        If the property is set to None, then the default ``transport`` will be obtained from :class:`.Loader`.
        """
        return self.__transport_factory

    @transport_factory.setter
    def transport_factory(self, new_transport_factory):
        from .transports import TransportFactory
        from ..loader import Loader

        self.__transport = None
        # If we don't set a new transport factory, then let's create a default one
        if new_transport_factory is None:
            default_transport = Loader.get('transport')
            self.__transport_factory = TransportFactory(default_transport)
        elif isinstance(new_transport_factory, TransportFactory):
            self.__transport_factory = new_transport_factory
        else:
            self.__transport_factory = TransportFactory(new_transport_factory)

    @property
    def appversion(self) -> str:
        """:return: The appversion defined at construction (used for creating requests by transports)
        :rtype: str"""
        return self.__appversion

    @property
    def user_agent(self) -> str:
        """:return: The user_agent defined at construction (used for creating requests by transports)
        :rtype: str"""
        return self.__user_agent

    @property
    def authenticated(self) -> bool:
        """:return: True if session is authenticated, False otherwise.
        :rtype: bool
        """
        return self.__UID is not None
    
    @property
    def UID(self) -> Optional[str]:
        """:return: the session UID, None if not authenticated
        :rtype: str, optional
        """
        return self.__UID

    @property
    def Scopes(self) -> Optional[list[str]]:
        """:return: list of scopes of the current session, None if unknown or not defined.
        :rtype: list[str], optional
        """
        return self.__Scopes

    @property
    def AccountName(self) -> str:
        """:return: session account name (mostly used for SSO)
        :rtype: str
        """
        return self.__AccountName

    @property
    def AccessToken(self) -> str:
        """:return: return the access token for API calls (used by transports)
        :rtype: str
        """
        return self.__AccessToken

    @property
    def needs_twofa(self) -> bool:
        """:return: True if a 2FA authentication is needed, False otherwise.
        :rtype: bool
        """
        if self.Scopes is None:
            return False
        return 'twofactor' in self.Scopes

    @property
    def supports_fido2(self) -> Optional[Fido2AssertionParameters]:
        """
        :return: If 2FA is needed, and FIDO2 is supported, return the
            parameters needed to create a FIDO2 assertion.
            None otherwise (either 2FA is not needed, or FIDO2 is not supported).
        :rtype: Fido2AssertionParameters, optional
        """
        if not self.__2FA:
            return None

        # When FIDO2 is not supported, the __2FA property contains:
        # {'AuthenticationOptions': None, 'RegisteredKeys': []}
        options = self.__2FA.get('FIDO2', {}).get("AuthenticationOptions")
        if not options:
            return None

        public_key = options.get("publicKey")
        if not public_key:
            return None

        return Fido2AssertionParameters(
            challenge=public_key["challenge"],
            rp_id=public_key["rpId"],
            allow_credentials=public_key["allowCredentials"],
            user_verification=public_key["userVerification"]
        )

    @property
    def environment(self):
        """Get/set the environment in use for that session. It can be only set once at the beginning of the session's object lifetime,
        as changing the environment can lead to security hole.

        If the new value is:
        
        * None: do nothing
        * a string: will use :meth:`Loader.get("environment", newvalue)` to get the actual class.
        * an environment: use it
        """
        if self.__environment is None:
            from proton.loader import Loader
            self.__environment = Loader.get('environment')()
        return self.__environment

    @environment.setter
    def environment(self, newvalue):
        # Do nothing if we set to None
        if newvalue is None:
            return
        if isinstance(newvalue, str):
            newvalue = Loader.get("environment", newvalue)()
        if not isinstance(newvalue, Environment):
            raise TypeError("environment should be a subclass of Environment")

        #Same environment => nothing to do
        if self.__environment == newvalue:
            return
        
        if self.__environment is not None:
            raise ValueError("Cannot change environment of an established session (that would create security issues)!")
        self.__environment = newvalue

    def __setstate__(self, data):
        # If we're running an unpickle, then the object constructor hasn't been called, so we need to populate __dict__

        # self.__appversion is the most basic attribute : if missing, so constructor has not been called
        is_init_called = "_Session__appversion" in self.__dict__
        if not is_init_called:
            for attr, default in (('gnupg_for_modulus', None), ('can_run_requests', None), ('transport', None), ('persistence_observers', [])):
                attrname = '_Session__' + attr
                self.__dict__[attrname] = default

        # Restore data from LastUseData if we don't have it already (allow pickle load)
        last_use_data = data.get('LastUseData', {})
        # `appversion` and `user_agent` are set in constructor (directly or via SSO class) : keep them as much as possible (in case of client version update, we want to have the correct version)
        if not is_init_called:
            for attr, default in (('appversion', 'Other'), ('user_agent', 'None')):
                attrname = '_Session__' + attr
                self.__dict__[attrname] = last_use_data.get(attr, default)
        # `2FA` and `refresh_revision` : need to be restored
        for attr, default in (('2FA', None), ('refresh_revision', 0)):
            attrname = '_Session__' + attr
            self.__dict__[attrname] = last_use_data.get(attr, default)
        
        # We don't pickle the transport, so if not set just use the default
        if '_Session__transport_factory' not in self.__dict__:
            self.transport_factory = None

        self.__UID = data.get('UID', None)
        self.__AccessToken = data.get('AccessToken', None)
        self.__RefreshToken = data.get('RefreshToken', None)
        self.__Scopes = data.get('Scopes', None)
        self.__AccountName = data.get('AccountName', None)
        #Reset transport (user agent etc might have changed)
        self.__transport = None
        #get environment as stored in the session
        if data.get('Environment', None) is not None:
            self.__environment: Environment = Loader.get("environment", data.get('Environment', None))()
        else:
            self.__environment = None

        # Store everything we don't know about in extrastate
        self.__extrastate = dict([(k, v) for k, v in data.items() if k not in ('UID','AccessToken','RefreshToken','Scopes','AccountName','Environment', 'LastUseData')])

    def __getstate__(self):
        # If we don't have an UID, then we're not logged in and we don't want to store a specific state
        if self.UID is None:
            data = {}
        else:
            data = {
                #Session data
                'UID': self.UID,
                'AccessToken': self.__AccessToken,
                'RefreshToken': self.__RefreshToken,
                'Scopes': self.Scopes,
                'Environment': self.environment.name,
                'AccountName': self.__AccountName,
                'LastUseData': {
                    '2FA': self.__2FA,
                    'appversion': self.__appversion,
                    'user_agent': self.__user_agent,
                    'refresh_revision': self.__refresh_revision,
                }
            }
            # Add the additional extra state data that we might have
            data.update(self.__extrastate)

        return data

    def _requests_lock(self, no_condition_check=False):
        """Lock the session, this has to be done when doing requests that affect the session state (i.e. :meth:`authenticate` for 
        instance), to prevent race conditions.

        Internally, this is done using :class:`asyncio.Event`.

        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        """
        if no_condition_check:
            return
        
        if self.__can_run_requests is None:
            self.__can_run_requests = asyncio.Event()
        self.__can_run_requests.clear()

        # Lock observers (we're about to modify the session)
        account_name = self.AccountName
        session_data = self.__getstate__()
        for observer in self.__persistence_observers:
            observer._acquire_session_lock(account_name, session_data)

    def _requests_unlock(self, no_condition_check=False, account_name=None):
        """Unlock the session, this has to be done after doing requests that affect the session state (i.e. :meth:`authenticate` for 
        instance), to prevent race conditions.

        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        :param account_name: Allow providing explicitly the account_name of the session, useful when it's for a logout when the session might not exist any more
        :type no_condition_check: str, optional
        """
        if no_condition_check:
            return
        
        if self.__can_run_requests is None:
            self.__can_run_requests = asyncio.Event()
        self.__can_run_requests.set()

        # Only store data if we have an actual account (session not logged in shouldn't store data)
        # If we have a known account, use it
        if self.AccountName is not None:
            account_name = self.AccountName
            session_data = self.__getstate__()
        else:
            session_data = None

        # Unlock observers (we might have modified the session)
        # It's important to do it in reverse order, as otherwise there's a risk of deadlocks
        for observer in reversed(self.__persistence_observers):
            observer._release_session_lock(account_name, session_data)

    async def _requests_wait(self, no_condition_check=False):
        """Wait for session unlock.

        :param no_condition_check: Internal flag to disable locking, defaults to False
        :type no_condition_check: bool, optional
        """
        if no_condition_check or self.__can_run_requests is None:
            return
        
        await self.__can_run_requests.wait()


    async def __sleep_for_exception(self, e):
        if e.http_headers.get('retry-after','-').isnumeric():
            await asyncio.sleep(int(e.http_headers.get('retry-after')))
        else:
            # No crypto risk here of using an unsafe generator
            await asyncio.sleep(3+random.random()*5)  # nosec B311 # nosemgrep: gitlab.bandit.B311

    async def __async_api_request_internal(
        self, endpoint,
        jsondata=None, data=None, additional_headers=None,
        method=None, params=None, no_condition_check=False,
        return_raw=False
    ):
        """Internal function to do an API request (without clever exception handling and retrying). 
        See :meth:`async_api_request` for the parameters specification."""
        # Should (and can we) create a transport
        if self.__transport is None and self.__transport_factory is not None:
            self.__transport = self.__transport_factory(self)
        if self.__transport is None:
            raise RuntimeError("Could not instanciate a transport, are required dependencies installed?")

        await self._requests_wait(no_condition_check)
        return await self.__transport.async_api_request(endpoint, jsondata, data, additional_headers, method, params,
                                                        return_raw=return_raw)

    def _verify_modulus(self, armored_modulus) -> bytes:
        if self.__gnupg_for_modulus is None:
            import gnupg
            # The modulus key is imported in a separate GPG keyring (rather than on the user's
            # default one) by modifying the gnupg home directory. The reason for doing this is
            # that we received crash reports due to users messing up with their default GPG
            # keyring permissions. In this case, python-gnupg fails silently when importing
            # Proton's modulus key, which then causes the signature verification errors.
            proton_gnupg_dir = Path(ExecutionEnvironment().path_cache) / "gnupg"
            proton_gnupg_dir.mkdir(exist_ok=True)
            self.__gnupg_for_modulus = gnupg.GPG(gnupghome=proton_gnupg_dir)
            self.__gnupg_for_modulus.import_keys(SRP_MODULUS_KEY)

        # gpg.decrypt verifies the signature too, and returns the parsed data.
        # By using gpg.verify the data is not returned
        verified = self.__gnupg_for_modulus.decrypt(armored_modulus)

        if not (verified.valid and verified.fingerprint.lower() == SRP_MODULUS_KEY_FINGERPRINT):
            raise ProtonCryptoError('Invalid modulus')

        return base64.b64decode(verified.data.strip())







"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from proton.session.formdata import FormData
from .. import Session
from ..exceptions import *
from .base import Transport, RawResponse

import json, base64, asyncio, aiohttp, hashlib
from OpenSSL import crypto
from typing import Iterable, Union, Optional

NOT_MODIFIED = 304

# It's stupid, but we have to inherit from aiohttp.Fingerprint to trigger the correct logic in aiohttp
class AiohttpCertkeyFingerprint(aiohttp.Fingerprint):
    def __init__(self, fingerprints: Optional[Iterable[Union[bytes, str]]]) -> None:
        if fingerprints is not None:
            self._fingerprints = []
            for fp in fingerprints:
                if type(fp) == str:
                    self._fingerprints.append(base64.b64decode(fp))
                else:
                    self._fingerprints.append(fp)
        else:
            self._fingerprints = None

    def check(self, transport: asyncio.Transport) -> None:
        if not transport.get_extra_info("sslcontext"):
            return
        # Can't check anything if we don't have fingerprints
        if self._fingerprints is None:
            return
        sslobj = transport.get_extra_info("ssl_object")
        cert = sslobj.getpeercert(binary_form=True)

        cert_obj = crypto.load_certificate(crypto.FILETYPE_ASN1, cert)
        pubkey_obj = cert_obj.get_pubkey()
        pubkey = crypto.dump_publickey(crypto.FILETYPE_ASN1, pubkey_obj)

        pubkey_hash = hashlib.sha256(pubkey).digest()

        if pubkey_hash not in self._fingerprints:
            # Dump certificate, so we can diagnose if needed with:
            # base64 -d|openssl x509 -text -inform DER 
            raise ProtonAPINotReachable(f"TLS pinning verification failed: {base64.b64encode(cert)}")


class AiohttpTransport(Transport):
    def __init__(self, session: Session, form_data_transformer: FormDataTransformer = None):
        super().__init__(session)
        self._form_data_transformer = form_data_transformer or FormDataTransformer()

    @classmethod
    def _get_priority(cls):
        return 10

    @property
    def tls_pinning_hashes(self):
        return self._environment.tls_pinning_hashes

    @property
    def http_base_url(self):
        return self._environment.http_base_url

    async def async_api_request(
        self, endpoint,
        jsondata=None, data=None, additional_headers=None,
        method=None, params=None, return_raw=False
    ) -> dict | RawResponse:
        if self.tls_pinning_hashes is not None:
            ssl_specs = AiohttpCertkeyFingerprint(self.tls_pinning_hashes)
        else:
            # Validate SSL normally if we didn't have fingerprints
            import ssl
            ssl_specs = ssl.create_default_context()
            ssl_specs.verify_mode = ssl.CERT_REQUIRED

        headers = {
            'x-pm-appversion': self._session.appversion,
            'User-Agent': self._session.user_agent,
        }
        if self._session.authenticated:
            headers['x-pm-uid'] = self._session.UID
            headers['Authorization'] = 'Bearer ' + self._session.AccessToken
        headers.update(self._environment.http_extra_headers)

        async with aiohttp.ClientSession(headers=headers) as s:
            # If we don't have an explicit method, default to get if there's no data, post otherwise
            if method is None:
                if not jsondata and not data:
                    fct = s.get
                else:
                    fct = s.post
            else:
                fct = {
                    'get': s.get,
                    'post': s.post,
                    'put': s.put,
                    'delete': s.delete,
                    'patch': s.patch
                }.get(method.lower())

                if fct is None:
                    raise ValueError("Unknown method: {}".format(method))

            form_data = self._form_data_transformer.to_aiohttp_form_data(data) if data else None

            try:
                async with fct(
                        self.http_base_url + endpoint, headers=additional_headers,
                        json=jsondata, data=form_data, params=params, ssl=ssl_specs
                ) as ret:
                    if return_raw:
                        return await self._build_raw(ret)

                    ret_json = await self._parse_json(ret)

                return ret_json
            except aiohttp.ClientError as e:
                raise ProtonAPINotReachable("Connection error.") from e
            except asyncio.TimeoutError as e:
                raise ProtonAPINotReachable("Timeout error.") from e
            except ProtonAPINotReachable:
                raise
            except ProtonAPIError:
                raise
            except Exception as e:
                raise ProtonAPIUnexpectedError(e)

    async def _build_raw(self, ret):
        response = RawResponse(ret.status,
                               tuple(ret.headers.items()), None, None)

        if ret.status == NOT_MODIFIED:
            return response

        response.data = await ret.read()

        if ret.headers['content-type'] != 'application/octet-stream':
            response.json = await self._parse_json(ret, allow_unmodified=True)

        return response

    async def _parse_json(self, ret, allow_unmodified=False):
        if ret.headers['content-type'] != 'application/json':
            raise ProtonAPINotReachable("API returned non-json results")
        try:
            ret_json = await ret.json()
        except json.decoder.JSONDecodeError:
            raise ProtonAPIError(ret.status, dict(ret.headers), {})

        if ret_json['Code'] not in [1000, 1001]:
            raise ProtonAPIError(ret.status, dict(ret.headers), ret_json)

        return ret_json


class FormDataTransformer:
    @staticmethod
    def to_aiohttp_form_data(form_data: FormData) -> aiohttp.FormData:
        """
        Converts proton.session.data.FormData into aiohttp.FormData.
        https://docs.aiohttp.org/en/stable/client_reference.html#formdata
        """
        result = aiohttp.FormData()
        for field in form_data.fields:
            result.add_field(
                name=field.name, value=field.value,
                content_type=field.content_type, filename=field.filename
            )
        return result



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from ._base import BasicView
import getpass
import sys

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ..session import Session


class BasicCLIView(BasicView):
    """Implementation of :class:`proton.views.BasicView` for a CLI. It's really just print + input calls."""
    def __init__(self):
        pass

    @classmethod
    def _get_priority(cls):
        return 0

    def display_error(self, message: str) -> None:
        print("Error: ", message, file=sys.stderr)

    def display_notice(self, message: str) -> None:
        print(message)

    def _session_to_string(self, s: "Session", default_session: "Session") -> str:
        flags = []
        if s == default_session:
            flags.append('default')
        if s.environment.name != 'prod':
            flags.append(f'env:{s.environment.name}')
        
        if len(flags) > 0:
            flags_str = f" [{', '.join(flags)}]"
        else:
            flags_str = ''
        return f'{s.AccountName}{flags_str}'

    def display_session_list(self, sessions : list["Session"], ask_to_select_one : bool = False) -> None:
        if len(sessions) == 0:
            print("No active sessions")
        else:
            print(f"Active session list [{len(sessions)}]:")
            print('')
            sorted_sessions = list(sorted(sessions, key=lambda x: x.AccountName))
            for session_id, s in enumerate(sorted_sessions):
                if ask_to_select_one:
                    print(f' [{session_id+1:2d}] {self._session_to_string(s, sessions[0])}')
                else:
                    print(f"- {self._session_to_string(s, sessions[0])}")

            if ask_to_select_one:
                while True:
                    user_input = input("Please select a session: ") # nosec (Python 3 only code)
                    if user_input.isnumeric():
                        user_input_idx = int(user_input) - 1
                        if user_input_idx >= 0 and user_input_idx < len(sorted_sessions):
                            return sorted_sessions[user_input_idx]
                        else: 
                            print("Invalid input!")
                    else:
                        for s in sorted_sessions:
                            if s.AccountName == user_input:
                                return s
                        print("Invalid input!")

    def ask_credentials(self, ask_login: bool = False, ask_password: bool = False, ask_2fa: bool = False) -> tuple[Optional[str], Optional[str], Optional[str]]:
        login = None
        password = None
        twofa = None
        if ask_login:
            login = input("Please enter your user name: ") # nosec (Python 3 only code)
            if login == '':
                login = None
        if ask_password:
            password = getpass.getpass()
            if password == '':  # nosec B105
                password = None  # nosec B105
        if ask_2fa:
            twofa = input("Please enter your 2FA code: ") # nosec (Python 3 only code)
            if twofa == '' or not twofa.isnumeric():
                twofa = None
        return login, password, twofa



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import weakref
from dataclasses import dataclass
from typing import Optional, Any, Tuple


@dataclass
class RawResponse:
    """
    A response that contains the status code and headers along with the body
    as json. This gives more context to clients when receiving a response.
    This type is returned where return_raw is set to True.

    :param status_code: The status code of the response
    :param headers: The headers in the response
    :param json: The response parsed as json
    :param data: The response as a binary blob of bytes
    """
    status_code: int
    headers: Tuple[Tuple[str, Any]]
    json: Optional[dict]
    data: Optional[bytes]

    def find_first_header(self, key, default=None):
        """
        Searches for the given key in the headers and returns the first value
        if found, otherwise returns the default value.
        """
        for k, v in self.headers:
            if key == k:
                return v
        return default


class Transport:
    """
    The base class of all transports. This class should be subclassed to
    implement the async_api_request method, which is the main method that
    is used to make requests to the API.

    A transport abstracts away the details of how requests are made to the API,
    for example, it could be using requests, aiohttp, or any other library.
    """
    def __init__(self, session):
        self.__session = weakref.ref(session)
    
    @property
    def _session(self):
        return self.__session()

    @property
    def _environment(self):
        #Shortcut to access environment
        return self._session.environment

    def __eq__(self, other):
        # It's the same transport if it's the same type (that's what users would generally assume)
        return self.__class__ == other.__class__

    async def is_working(self):
        try:
            return await self.async_api_request('/tests/ping').get('Code') == '1000'
        except:
            return False

    async def async_api_request(
        self, endpoint,
        jsondata=None, additional_headers=None,
        method=None, params=None
    ):
        raise NotImplementedError("async_api_request should be implemented")

class TransportFactory:
    def __init__(self, cls, *args, **kwargs):
        self._cls = cls
        self._args = args
        self._kwargs = kwargs

    def __call__(self, session):
        return self._cls(session, *self._args, **self._kwargs)

    def __eq__(self, other):
        # It's the same transport if it's the same type (that's what users would generally assume)
        return self._cls == other._cls



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

import base64
import fcntl
import os
import re
from typing import TYPE_CHECKING, Optional

from proton.keyring import Keyring

if TYPE_CHECKING:
    from ..session import Session


# We don't necessarily need it to be a singleton, it doesn't harm in itself if multiple instances are created
class ProtonSSO:
    """Proton Single Sign On implementation. This allows session persistence for the current user.

    The general approach for this is to create a SSO instance, and then to get either a specific or the default session, and work from there:

    .. code-block::

        from proton.sso import ProtonSSO
        sso = ProtonSSO()
        session = sso.get_default_session()
        # or:
        session = sso.get_session('pro') # get session for account pro

    Note that it is advised not to try to "guess" the state of the session, but instead to just try to use it, and handle any exception that would arise.

    This object uses advisory locks (using ``flock``) to protect the session from multiple conflicting changes. This does not guarantee that
    Session objects are immune to what happens in another process (i.e. imagine if one process terminates the session.), but at least makes it consistent.
    In the future, it would be nice to use an IPC mechanism to make sure other processes are aware of the state change.
    """
    def __init__(self, appversion : str = "Other", user_agent: str ="None", keyring_backend_name=None):
        """Create a SSO instance

        :param appversion: Application version (see :class:`proton.session.Session`), defaults to "Other"
        :type appversion: str, optional
        :param user_agent: User agent version (see :class:`proton.session.Session`), defaults to "None"
        :type user_agent: str, optional
        :param keyring_backend_name: Name of Keyring backend to load (see :class:`proton.keyring.Keyring`), defaults to None
        :type keyring_backend_name: str, optional
        """
        # Store appversion and user_agent for subsequent sessions
        self._appversion = appversion
        self._user_agent = user_agent

        from ..utils import ExecutionEnvironment
        self._adv_locks_path = ExecutionEnvironment().path_runtime
        self._adv_locks = {}

        self._session_data_cache = {}

        # This is a global lock, we use it when we modify the indexes
        self._global_adv_lock = open(os.path.join(self._adv_locks_path, "proton-sso.lock"), "w", encoding="utf-8")  # nosemgrep: python.lang.best-practice.open-never-closed.open-never-closed
        self.__keyring_backend = None
        self.__keyring_backend_name = keyring_backend_name

    def __encode_name(self, account_name) -> str:
        """Helper function to convert an account_name into a safe alphanumeric string.

        :param account_name: normalized account_name
        :type account_name: str
        :return: base32 encoded string, without padding.
        :rtype: str
        """
        return base64.b32encode(account_name.encode('utf8')).decode('ascii').rstrip('=').lower()

    def __keyring_key_name(self, account_name : str) -> str:
        """Helper function to get the keyring key for account_name

        :param account_name: normalized account_name
        :type account_name: str
        :return: keyring key
        :rtype: str
        """
        return f'proton-sso-account-{self.__encode_name(account_name)}'

    def __keyring_index_name(self) -> str:
        """Helper function to get the keyring key to store the index (i.e. account names in order)

        :return: keyring key
        :rtype: str
        """
        return f'proton-sso-accounts'

    @property
    def _keyring(self) -> "Keyring":
        """Shortcut to get the default keyring backend

        :return: an instance of the default Keyring
        :rtype: Keyring
        """
        if self.__keyring_backend is None:
            self.__keyring_backend = Keyring.get_from_factory(self.__keyring_backend_name)
        elif not isinstance(self.__keyring_backend, type(Keyring.get_from_factory(self.__keyring_backend_name))):
            # If the current keyring does not match the keyring we were using previously,
            # then something must've changed in the env and we should raise an exception.
            raise RuntimeError("Keyring backends do not match")

        return self.__keyring_backend

    @property
    def sessions(self) -> list[str]:
        """Returns the account names for the current system user

        :return: list of normalized account_names
        :rtype: list[str]
        """

        # We might remove invalid session and clean the index, so create a full lock on the SSO object
        fcntl.flock(self._global_adv_lock, fcntl.LOCK_EX)
        try:
            keyring = self._keyring

            try:
                keyring_index = keyring[self.__keyring_index_name()]
            except KeyError:
                keyring_index = []

            cleaned_index = [account_name for account_name in keyring_index if len(self._get_session_data(account_name)) > 0]
            if cleaned_index != keyring_index:
                keyring[self.__keyring_index_name()] = cleaned_index

            # Try to remove any account from keyring that we've removed from SSO
            for removed_account in set(keyring_index).difference(cleaned_index):
                try:
                    del keyring[self.__keyring_key_name(removed_account)]
                except KeyError:
                    pass

            return cleaned_index
        finally:
            fcntl.flock(self._global_adv_lock, fcntl.LOCK_UN)

    def get_session(self, account_name : Optional[str], override_class : Optional[type] = None) -> "Session":
        """Get the session identified by account_name

        :param account_name: account name to use. If None will return an empty session (can be used as a factory)
        :type account_name: Optional[str]
        :param override_class: Class to use for the session to be returned, by default will use proton.session.Session
        :type override_class: Optional[type]
        :return: the Session object. It will be an empty session if there's no session for account_name
        :rtype: Session
        """
        from ..session import Session

        if override_class is None:
            override_class = Session

        session = override_class(self._appversion, self._user_agent)
        session.register_persistence_observer(self)

        # If we have an account, then let's fetch the data from it. Otherwise we just ignore and return a blank session
        if account_name is not None:
            try:
                session_data = self._get_session_data(account_name)
            except KeyError:
                session_data = None
        else:
            session_data = None

        if session_data is not None:
            session.__setstate__(session_data)

        return session

    def get_default_session(self, override_class : Optional[type] = None)  -> "Session":
        """Get the default session for the system user. It will always be one valid session if one exists.

        :param override_class: Class to use for the session to be returned, see :meth:`get_session`.
        :type override_class: Optional[type]
        :return: the Session object. It will be an empty session if there's no session at all
        :rtype: Session
        """
        sessions = self.sessions
        if len(sessions) == 0:
            account_name = None
        else:
            account_name = sessions[0]

        return self.get_session(account_name, override_class)

    def set_default_account(self, account_name : str):
        """Set the default account for user to be account_name

        :param account_name: the account_name to use as default
        :type account_name: str
        :raises KeyError: if the account name is unknown
        """
        # We might be reordering accounts, so let's lock the full sso so we can't have concurrent actions here
        fcntl.flock(self._global_adv_lock, fcntl.LOCK_EX)
        try:
            keyring = self._keyring

            try:
                keyring_index = keyring[self.__keyring_index_name()]
            except KeyError:
                keyring_index = []

            if account_name not in keyring_index:
                raise KeyError(account_name)

            new_keyring_index = [account_name] + [x for x in keyring_index if x != account_name]
            if new_keyring_index != keyring_index:
                keyring[self.__keyring_index_name()] = new_keyring_index
            
        finally:
            fcntl.flock(self._global_adv_lock, fcntl.LOCK_UN)

    def _get_session_data(self, account_name : str) -> dict:
        """Helper function to get data of a session, returns an empty dict if no data is present

        :param account_name: normalized account name
        :type account_name: str
        :return: content of the session data, empty dict if it doesn't exist.
        :rtype: dict
        """
        try:
            data = self._keyring[self.__keyring_key_name(account_name)]
        except KeyError:
            data = {}

        # This is an encapsulation violation (we're not supposed to know that the account name is stored in AccountName)
        # It allows us nevertheless to validate that the session contains actual data, which is good to not break if a
        # Session implementation is invalid.
        if data.get('AccountName') != account_name:
            data = {}

        return data


    def _acquire_session_lock(self, account_name : str, current_data : dict) -> None:
        """Observer pattern for :class:`proton.session.Session` (see :meth:`proton.session.Session.register_persistence_observer`). It is called when the Session object is getting locked, because it's expected to be changed
        and we want to avoid race conditions.

        :param account_name: account name of the session
        :type account_name: str
        :param current_data: current session data serialized as a dictionary
        :type current_data: dict
        """
        if not account_name:
            # Don't do anything, we don't know the account yet!
            return

        self._adv_locks[account_name] = open(  # nosemgrep: python.lang.best-practice.open-never-closed.open-never-closed # pylint: disable=line-too-long
            os.path.join(self._adv_locks_path, f'proton-sso-{self.__encode_name(account_name)}.lock'),
            "w", encoding="utf-8"
        )
        # This is a blocking call. 
        # FIXME: this is Linux specific
        fcntl.flock(self._adv_locks[account_name], fcntl.LOCK_EX)

        self._session_data_cache[account_name] = current_data


    def _release_session_lock(self, account_name : str, new_data : dict) -> None:
        """Observer pattern for :class:`proton.session.Session` (see :meth:`proton.session.Session.register_persistence_observer`). It is called when the Session object is getting unlocked.

        If the data between has changed since :meth:`_acquire_session_lock` was called, it will be persisted in the keyring.

        :param account_name: account name of the session
        :type account_name: str
        :param new_data: current session data serialized as a dictionary
        :type new_data: dict
        """
        if not account_name:
            # Don't do anything, we don't know the account yet!
            return

        if new_data is not None and len(new_data) > 0 and new_data.get('AccountName', None) != account_name:
            raise ValueError("Sessions need to store a valid AccountName in order to store data.")

        # Don't do anything if data hasn't changed
        if account_name in self._session_data_cache:
            if self._session_data_cache[account_name] == new_data:
                return
            del self._session_data_cache[account_name]

        # We might be reordering accounts, so let's lock the full sso so we can't have concurrent actions here
        fcntl.flock(self._global_adv_lock, fcntl.LOCK_EX)
        try:
            keyring = self._keyring

            # Get current data
            try:
                keyring_entry = keyring[self.__keyring_key_name(account_name)]
            except KeyError:
                keyring_entry = {}

            try:
                keyring_index = keyring[self.__keyring_index_name()]
            except KeyError:
                keyring_index = []

            # By default, we don't change anything
            new_keyring_index = keyring_index

            # No data, this is probably a logout
            if new_data is None or len(new_data) == 0:
                # Discard from the index
                new_keyring_index = [x for x in keyring_index if x != account_name]

                # Delete the entry if we had some data previously
                if len(keyring_entry) > 0:
                    del keyring[self.__keyring_key_name(account_name)]
            # We have new data
            else:
                # If this is a new entry, then append the index with the account (we leave the default as is)
                if account_name not in keyring_index:
                    new_keyring_index = keyring_index + [account_name]

                # Store the new data
                keyring[self.__keyring_key_name(account_name)] = new_data

            # We only store the new index if it has changed (wouldn't harm to do it anyway)
            if new_keyring_index != keyring_index:
                keyring[self.__keyring_index_name()] = new_keyring_index

        finally:
            fcntl.flock(self._global_adv_lock, fcntl.LOCK_UN)

        if account_name in self._adv_locks:
            # FIXME: this is Linux specific
            fcntl.flock(self._adv_locks[account_name], fcntl.LOCK_UN)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import json
import os

from ..utils import ExecutionEnvironment
from ._base import Keyring
from .exceptions import KeyringError


class KeyringBackendJsonFiles(Keyring):
    """Primitive data storage implementation, to be used when no better keyring is present.

    It stores each entry a json in the configuration path.
    """
    def __init__(self, path_config=None):
        super().__init__()
        self.__path_base = path_config or ExecutionEnvironment().path_config

    def _get_item(self, key):
        filepath = self.__get_filename_for_key(key)
        if not os.path.exists(filepath):
            raise KeyError(key)

        try:
            with open(filepath, "r", encoding="utf-8") as f:
                return json.load(f)
        except json.JSONDecodeError as e:
            self._del_item(key)
            raise KeyError(key) from e

    def _del_item(self, key):
        filepath = self.__get_filename_for_key(key)
        if not os.path.exists(filepath):
            raise KeyError(key)

        os.unlink(filepath)

    def _set_item(self, key, value):
        try:
            with open(self.__get_filename_for_key(key), "w", encoding="utf-8") as f:
                json.dump(value, f)
        except TypeError as e:
            # The value we got is not serializable, thus a type error is thrown,
            # we re-raise it as a ValueError because the value that was provided was in
            # in un-expected format/type
            raise ValueError(value) from e
        except FileNotFoundError as e:
            # if the path was not previously created for some reason,
            # we get a FileNotFoundError
            raise KeyringError(key) from e

    def __get_filename_for_key(self, key):
        return os.path.join(self.__path_base, f'keyring-{key}.json')

    @classmethod
    def _get_priority(cls) -> int:
        return -1000

    @classmethod
    def _validate(cls):
        is_able_to_write_in_dir = True
        try:
            ExecutionEnvironment().path_config
        except: # noqa
            is_able_to_write_in_dir = False

        return is_able_to_write_in_dir



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from .metaclasses import Singleton
import os
import shutil
# Try to get the BaseDirectory module
try:
    from xdg import BaseDirectory
except ImportError:
    BaseDirectory = None


class ExecutionEnvironment(metaclass=Singleton):
    PROTON_DIR_NAME = "Proton"

    def __init__(self):
        # If we run as a system user, use system paths
        if os.getuid() == 0:
            self._setup_as_system_user()
        else:
        # If we run as a normal user
            self._setup_as_regular_user()

    @property
    def path_config(self):
        self.generate_dirs(self._path_config)
        return self._path_config

    @property
    def path_cache(self):
        self.generate_dirs(self._path_cache)
        return self._path_cache

    @property
    def path_logs(self):
        self.generate_dirs(self._path_logs)
        return self._path_logs

    @property
    def path_runtime(self):
        self.generate_dirs(self._path_runtime)
        return self._path_runtime

    @property
    def systemd_unit(self):
        return self._path_systemd_unit

    def generate_dirs(self, path):
        if os.path.isdir(path):
            return

        os.makedirs(path, mode=0o700, exist_ok=True)

    def _setup_as_system_user(self):
        self._path_config = f'/etc/{self.PROTON_DIR_NAME}'
        self._path_cache = f'/var/cache/{self.PROTON_DIR_NAME}'
        self._path_logs = f'/var/log/{self.PROTON_DIR_NAME}'
        self._path_runtime = f'/run/{self.PROTON_DIR_NAME}'
        self._path_systemd_unit = '/etc/systemd/system'

    def _setup_as_regular_user(self):
        config_home, cache_home, runtime_dir = self._get_dir_paths()

        self._path_config = os.path.join(config_home, self.PROTON_DIR_NAME)
        self._path_cache = os.path.join(cache_home, self.PROTON_DIR_NAME)
        self._path_logs = os.path.join(cache_home, self.PROTON_DIR_NAME, 'logs')
        self._path_runtime = os.path.join(runtime_dir, self.PROTON_DIR_NAME)
        self._path_systemd_unit = os.path.join(config_home, "systemd", "user")

    def _get_dir_paths(self):
        # If BaseDirectory is found then we can extract valuable data from it
        if BaseDirectory:
            config_home = BaseDirectory.xdg_config_home
            cache_home = BaseDirectory.xdg_cache_home
            runtime_dir = BaseDirectory.get_runtime_dir()
        else:
        # Otherwise use default constructed from $HOME environment variable
            home = os.environ.get('HOME', None)
            if home is None:
                raise RuntimeError("Cannot figure out where to place files, is $HOME defined?")

            config_home = os.path.join(home, '.config')
            cache_home = os.path.join(home, '.cache')
            runtime_dir = f'/run/user/{os.getuid()}'

        return config_home, cache_home, runtime_dir

class ProductExecutionEnvironment(ExecutionEnvironment):
    """
    This class serves the purpose of helping in standardizing folder structure
    across products. Thus each product should derive from `ProductExecutionEnvironment`
    and setting the class property `PRODUCT` to match its correspondent product.

    This should help to more easily find files and improving cross-product
    collaboration. 
    """
    PRODUCT = None

    def __init__(self):
        super().__init__()
        if self.PRODUCT is None:
            raise RuntimeError("`PRODUCT` is not set")

    @property
    def path_config(self):
        path = os.path.join(super().path_config, self.PRODUCT)
        self.generate_dirs(path)
        return path

    @property
    def path_cache(self):
        path = os.path.join(super().path_cache, self.PRODUCT)
        self.generate_dirs(path)
        return path

    @property
    def path_logs(self):
        path = os.path.join(super().path_logs, self.PRODUCT)
        self.generate_dirs(path)
        return path

    @property
    def path_runtime(self):
        path = os.path.join(super().path_runtime, self.PRODUCT)
        self.generate_dirs(path)
        return path



class VPNExecutionEnvironment(ProductExecutionEnvironment):
    """Execution environment dedicated for the VPN product."""
    PRODUCT = "VPN"



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import re
from typing import Union
from proton.loader import Loader


class Keyring:
    """Base class for keyring implementations.

    Keyrings emulate a dictionary, with:

    * keys: lower case alphanumeric strings (dashes are allowed)
    * values: JSON-serializable list or dictionary.
    """
    def __init__(self):
        pass

    @classmethod
    def get_from_factory(cls, backend: str = None) -> "Keyring":
        """
            :param backend: Optional.
                Specific backend name.

        If backend is passed then it will attempt to get that specific
        backend, otherwise it will attempt to get the default backend.
        The definition of default is as follows:

         - The backend passes the `_validate()`
         - The backend with the highest `_get_priority()` value
        :raises RuntimeError: if there's no available backend
        """
        keyring_backend = Loader.get("keyring", class_name=backend)

        return keyring_backend()

    def __getitem__(self, key: str):
        """Get an item from the keyring

        :param key: Key (lowercaps alphanumeric, dashes are allowed)
        :type key: str
        :raises TypeError: if key is not of valid type
        :raises ValueError: if key doesn't satisfy constraints
        :raises KeyError: if key does not exist
        :raises KeyringLocked: if keyring is locked when it shouldn't be
        :raises KeyringError: if there's something broken with keyring
        """
        self._ensure_key_is_valid(key)
        return self._get_item(key)

    def __delitem__(self, key: str):
        """Remove an item from the keyring

        :param key: Key (lowercaps alphanumeric, dashes are allowed)
        :type key: str
        :raises TypeError: if key is not of valid type
        :raises ValueError: if key doesn't satisfy constraints
        :raises KeyError: if key does not exist
        :raises KeyringLocked: if keyring is locked when it shouldn't be
        :raises KeyringError: if there's something broken with keyring
        """
        self._ensure_key_is_valid(key)
        self._del_item(key)

    def __setitem__(self, key: str, value: Union[dict, list]):
        """Add or replace an item in the keyring

        :param key: Key (lowercaps alphanumeric, dashes are allowed)
        :type key: str
        :param value: Value to set. It has to be json-serializable.
        :type value: dict or list
        :raises TypeError: if key or value is not of valid type
        :raises ValueError: if key or value doesn't satisfy constraints
        :raises KeyringLocked: if keyring is locked when it shouldn't be
        :raises KeyringError: if there's something broken with keyring
        """
        self._ensure_key_is_valid(key)
        self._ensure_value_is_valid(value)
        self._set_item(key, value)

    def _get_item(self, key: str):
        raise NotImplementedError

    def _del_item(self, key: str):
        raise NotImplementedError

    def _set_item(self, key: str, value: Union[dict, list]):
        raise NotImplementedError

    def _ensure_key_is_valid(self, key):
        """Ensure key satisfies requirements"""
        if type(key) != str:
            raise TypeError(f"Invalid key for keyring: {key!r}")
        if not re.match(r'^[a-z0-9-]+$', key):
            raise ValueError("Keyring key should be alphanumeric")

    def _ensure_value_is_valid(self, value):
        """Ensure value satisfies requirements"""
        if not isinstance(value, dict) and not isinstance(value, list):
            raise TypeError(f"Provided value {value} is not a valid type (expect dict or list)")

    @classmethod
    def _get_priority(cls) -> int:
        return None

    @classmethod
    def _validate(cls):
        return False



"""
Copyright (c) 2023 Proton AG

This file is part of Proton.

Proton is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc
from typing import Union, Optional

class Environment(metaclass=abc.ABCMeta):
    @property
    def name(cls):
        cls_name = cls.__class__.__name__
        # dev should ensure that to avoid issues
        assert cls_name.endswith('Environment'), "Incorrectly named class"  # nosec B101 # nosemgrep: gitlab.bandit.B101 # noqa: E501 # pylint: disable=line-too-long
        return cls_name[:-11].lower()

    @property
    def http_extra_headers(self):
        #This can be overriden, but by default we don't add extra headers
        return {}

    @property
    @abc.abstractmethod
    def http_base_url(self):
        pass

    @property
    @abc.abstractmethod
    def tls_pinning_hashes(self):
        pass

    @property
    @abc.abstractmethod
    def tls_pinning_hashes_ar(self):
        pass

    def __eq__(self, other):
        if other is None:
            return False
        return self.name == other.name



class ProdEnvironment(Environment):
    @classmethod
    def _get_priority(cls):
        return 10

    @property
    def http_base_url(self):
        return "https://vpn-api.proton.me"

    @property
    def tls_pinning_hashes(self):
        return set([
            "CT56BhOTmj5ZIPgb/xD5mH8rY3BLo/MlhP7oPyJUEDo=",
            "35Dx28/uzN3LeltkCBQ8RHK0tlNSa2kCpCRGNp34Gxc=",
            "qYIukVc63DEITct8sFT7ebIq5qsWmuscaIKeJx+5J5A=",
        ])

    @property
    def tls_pinning_hashes_ar(self):
        return set([
            "EU6TS9MO0L/GsDHvVc9D5fChYLNy5JdGYpJw0ccgetM=",
            "iKPIHPnDNqdkvOnTClQ8zQAIKG0XavaPkcEo0LBAABA=",
            "MSlVrBCdL0hKyczvgYVSRNm88RicyY04Q2y5qrBt0xA=",
            "C2UxW0T1Ckl9s+8cXfjXxlEqwAfPM4HiW2y3UdtBeCw="
        ])
